from typing import Literal
from jaxtyping import Shaped, Float, Inexact
from torch import Tensor

from functools import partial
from math import prod, ceil

import torch

try:
    import triton
    import triton.language as tl

    TRITON_ENABLED = True
except ImportError:
    from torchlinops.utils import fake_triton as triton, fake_tl as tl

    TRITON_ENABLED = False

from .kernels import (
    weights1d,
    weights2d,
    weights3d,
    weights_torch,
    get_kernel_fn,
    _apply_default_kernel_params,
    KernelTypeStr,
    mod_pos,
)
from ._batch import batch_iterator

__all__ = ["ungrid"]

# Limit maximum number of grids to launch
TRITON_MAX_GRID_SIZE = 2**16 - 1
# Limit maximum kernel width in each dimension
TRITON_MAX_KERNEL_WIDTH_1D = 2**6


def ungrid(
    vals: Inexact[Tensor, "..."],
    locs: Float[Tensor, "... D"],
    width: float | tuple[float, ...],
    kernel: str = "kaiser_bessel",
    norm: str = "1",
    pad_mode: Literal["zero", "circular"] = "circular",
    kernel_params: dict = None,
):
    """Interpolate from on-grid values to off-grid locations.

    norm: str, 1 or 2
        if 2, uses Euclidean norm to grid points to compute weights
        if 1, computes weights as product of axis-aligned norm weights
            - Same as sigpy
    """

    kernel_params = {} if kernel_params is None else kernel_params
    vals_flat, locs, shapes = prep_ungrid_shapes(vals, locs, width)
    kernel_params = _apply_default_kernel_params(kernel, kernel_params)
    out_flat = _ungrid(
        vals_flat,
        locs,
        kernel=kernel,
        norm=norm,
        pad_mode=pad_mode,
        kernel_params=kernel_params,
        **shapes,
    )
    out = out_flat.reshape(*shapes["batch_shape"], *shapes["locs_batch_shape"])
    return out


def _ungrid(
    vals: Shaped[Tensor, "B ..."],
    locs: Float[Tensor, "... D"],
    grid_size: tuple[int, ...],
    width: tuple[float, ...],
    kernel: KernelTypeStr,
    norm: str,
    pad_mode: str,
    ndim: int,
    nbatch: int,
    is_complex: bool,
    locs_batch_shape: tuple[int, ...],
    npts: int,
    kernel_params,
    **kwargs,
):
    if vals.is_cuda and ndim in UNGRID.keys():
        # Ensure contiguity
        vals = vals.contiguous()
        locs = locs.contiguous()
        # Preallocate output
        output = torch.zeros(
            nbatch,
            *locs_batch_shape,
            dtype=vals.dtype,
            device=vals.device,
        )
        if is_complex:
            vals = torch.view_as_real(vals).contiguous()
            output = torch.view_as_real(output).contiguous()
        grid = _get_grid()  # TODO
        BLOCK_WIDTH = get_block_width(width, ndim, is_complex)
        with torch.cuda.device(vals.device):
            UNGRID[ndim][grid](
                vals,
                locs,
                output,
                nbatch,
                npts,
                kernel,
                norm,
                pad_mode,
                is_complex,
                *grid_size,
                *width,
                *BLOCK_WIDTH,
                **kernel_params,
            )
        if is_complex:
            output = torch.view_as_complex(output)
    else:
        # Pytorch fallback
        output = ungrid_torch(
            vals,
            locs,
            grid_size,
            width,
            kernel,
            norm,
            pad_mode,
            **kernel_params,
        )
    return output


def _get_grid():
    grid = lambda meta: (ceil(meta["npts"] / meta["pts_per_grid"]) * meta["nbatch"],)  # noqa: E731
    return grid


def get_block_width(kernel_width: tuple[float, ...], ndim: int, is_complex: bool):
    """Get necessary block width based on dimension and dtype of input"""
    block_width = list(triton.next_power_of_2(ceil(w + 1)) for w in kernel_width)
    test_block_width = block_width[:]  # Shallow copy
    if is_complex:
        test_block_width[-1] *= 2
    if max(test_block_width) > TRITON_MAX_KERNEL_WIDTH_1D:
        raise ValueError(
            f"Necessary block width {test_block_width} has entry which exceeds maximum width {TRITON_MAX_KERNEL_WIDTH_1D} (kernel width is doubled in last dim if input is complex)"
        )

    return block_width


@triton.heuristics(
    values={
        "pts_per_grid": lambda args: max(
            1, triton.cdiv(args["npts"] * args["nbatch"], TRITON_MAX_GRID_SIZE)
        ),
    },
)
@triton.jit
def _ungrid1d(
    in_ptr,
    pts_ptr,
    out_ptr,
    nbatch,
    npts,
    KERNEL: tl.constexpr,
    NORM: tl.constexpr,
    PAD_MODE: tl.constexpr,
    is_complex: tl.constexpr,  # bool
    x_size,
    x_kernel_width,
    X_BLOCK_WIDTH: tl.constexpr,
    pts_per_grid,  # Determined via heuristic
    beta=1.0,  # For kernel=kaiser_bessel
):
    """
    NORM has no effect in 1d
    """
    size = x_size

    pid_0 = tl.program_id(0)
    grids_per_batch = tl.cast(tl.ceil(npts / pts_per_grid), tl.int32)
    N, grid_start = pid_0 // grids_per_batch, pid_0 % grids_per_batch
    pts_lower, pts_upper = pts_per_grid * grid_start, pts_per_grid * (grid_start + 1)

    x_base_range = tl.arange(0, X_BLOCK_WIDTH)
    if is_complex:
        # Last dimension has double size because (real, imag) are interleaved
        # Overall size doubles as a result
        size = 2 * size

    in_batch_offset = N * size
    out_batch_offset = N * npts

    for p in range(pts_lower, pts_upper):
        if p < npts:
            x_target = tl.load(pts_ptr + p)
            weights, x_range, x_mask = weights1d(
                x_target, x_kernel_width, x_base_range, KERNEL, beta
            )

            if is_complex:
                # Pytorch interleaved indexing
                # Only applies to last dimension
                x_range_real = 2 * x_range
                x_range_imag = 2 * x_range + 1
                x_range_cplx = tl.join(x_range_real, x_range_imag)  # [width, 2]
                x_mask_cplx = tl.join(x_mask, x_mask)
                if PAD_MODE == "zero":
                    x_mask_cplx &= (x_range_cplx >= 0) & (x_range_cplx < (2 * x_size))
                elif PAD_MODE == "circular":
                    x_range_cplx = mod_pos(x_range_cplx, 2 * x_size)

                # Load
                grid_cplx = tl.load(
                    in_ptr + in_batch_offset + x_range_cplx, x_mask_cplx
                )
                grid_mask_cplx = x_mask_cplx
                # Split and process separately
                grid_real, grid_imag = tl.split(grid_cplx)
                mask_real, mask_imag = tl.split(grid_mask_cplx)
                out_real = tl.sum(weights * grid_real * mask_real)
                out_imag = tl.sum(weights * grid_imag * mask_imag)
                tl.store(out_ptr + 2 * (out_batch_offset + p), out_real)
                tl.store(out_ptr + 2 * (out_batch_offset + p) + 1, out_imag)

            else:
                # Normal indexing
                # x_range = x_base_range + x_nbhd
                if PAD_MODE == "zero":
                    x_mask &= (x_range >= 0) & (x_range < x_size)
                elif PAD_MODE == "circular":
                    x_range = mod_pos(x_range, x_size)

                # Load
                grid = tl.load(in_ptr + in_batch_offset + x_range, x_mask)

                # Process jointly
                out = tl.sum(weights * grid * x_mask)
                tl.store(out_ptr + out_batch_offset + p, out)


@triton.heuristics(
    values={
        "pts_per_grid": lambda args: max(
            1, triton.cdiv(args["npts"] * args["nbatch"], TRITON_MAX_GRID_SIZE)
        ),
    },
)
@triton.jit
def _ungrid2d(
    in_ptr,
    pts_ptr,
    out_ptr,
    nbatch,
    npts,
    KERNEL: tl.constexpr,
    NORM: tl.constexpr,
    PAD_MODE: tl.constexpr,
    is_complex: tl.constexpr,  # bool
    # Size of grid
    x_size,
    y_size,
    # Size of kernel
    x_kernel_width,
    y_kernel_width,
    # Size of blocks to load
    X_BLOCK_WIDTH: tl.constexpr,
    Y_BLOCK_WIDTH: tl.constexpr,
    pts_per_grid,  # Determined via heuristic
    beta=1.0,  # For kernel=kaiser_bessel
):
    """ """
    size = x_size * y_size

    pid_0 = tl.program_id(0)
    grids_per_batch = tl.cast(tl.ceil(npts / pts_per_grid), tl.int32)
    N, grid_start = pid_0 // grids_per_batch, pid_0 % grids_per_batch
    pts_lower, pts_upper = pts_per_grid * grid_start, pts_per_grid * (grid_start + 1)

    x_base_range = tl.arange(0, X_BLOCK_WIDTH)
    y_base_range = tl.arange(0, Y_BLOCK_WIDTH)
    if is_complex:
        # Last dimension has double size because (real, imag) are interleaved
        # Overall size doubles as a result
        size = 2 * size

    in_batch_offset = N * size
    out_batch_offset = N * npts

    for p in range(pts_lower, pts_upper):
        if p < npts:
            # Load target point
            x_target = tl.load(pts_ptr + 2 * p)
            y_target = tl.load(pts_ptr + 2 * p + 1)

            weights, x_range, y_range, x_mask, y_mask = weights2d(
                x_target,
                y_target,
                x_kernel_width,
                y_kernel_width,
                x_base_range,
                y_base_range,
                KERNEL,
                NORM,
                beta,
            )

            if is_complex:
                x_range_cplx = x_range  # [width]
                x_mask_cplx = x_mask
                # Pytorch interleaved indexing
                # Only applies to last dimension
                y_range_real = 2 * y_range
                y_range_imag = 2 * y_range + 1
                y_range_cplx = tl.join(y_range_real, y_range_imag)  # [width, 2]
                y_mask_cplx = tl.join(y_mask, y_mask)
                if PAD_MODE == "zero":
                    x_mask_cplx &= (x_range_cplx >= 0) & (x_range_cplx < x_size)
                    y_mask_cplx &= (y_range_cplx >= 0) & (y_range_cplx < (2 * y_size))
                elif PAD_MODE == "circular":
                    x_range_cplx = mod_pos(x_range_cplx, x_size)
                    y_range_cplx = mod_pos(y_range_cplx, 2 * y_size)

                grid_range_cplx = (
                    x_range_cplx[:, None, None] * y_size * 2 + y_range_cplx[None, :, :]
                )
                grid_mask_cplx = x_mask_cplx[:, None, None] & y_mask_cplx[None, :, :]

                # Load
                grid_cplx = tl.load(
                    in_ptr + in_batch_offset + grid_range_cplx, grid_mask_cplx
                )
                # Split and process separately
                grid_real, grid_imag = tl.split(grid_cplx)
                mask_real, mask_imag = tl.split(grid_mask_cplx)
                out_real = tl.sum(weights * grid_real)
                out_imag = tl.sum(weights * grid_imag)
                tl.store(out_ptr + 2 * (out_batch_offset + p), out_real)
                tl.store(out_ptr + 2 * (out_batch_offset + p) + 1, out_imag)

            else:
                # Normal indexing
                if PAD_MODE == "zero":
                    x_mask &= (x_range >= 0) & (x_range < x_size)
                    y_mask &= (y_range >= 0) & (y_range < y_size)
                elif PAD_MODE == "circular":
                    x_range = mod_pos(x_range, x_size)
                    y_range = mod_pos(y_range, y_size)

                grid_range = x_range[:, None] * y_size + y_range[None, :]
                grid_mask = x_mask[:, None] & y_mask[None, :]

                # Load
                grid = tl.load(in_ptr + in_batch_offset + grid_range, grid_mask)

                # Process jointly
                out = tl.sum(weights * grid)
                tl.store(out_ptr + out_batch_offset + p, out)


@triton.heuristics(
    values={
        "pts_per_grid": lambda args: max(
            1, triton.cdiv(args["npts"] * args["nbatch"], TRITON_MAX_GRID_SIZE)
        ),
    },
)
@triton.jit
def _ungrid3d(
    in_ptr,
    pts_ptr,
    out_ptr,
    nbatch,
    npts,
    KERNEL: tl.constexpr,
    NORM: tl.constexpr,
    PAD_MODE: tl.constexpr,
    is_complex: tl.constexpr,  # bool
    # Size of grid
    x_size,
    y_size,
    z_size,
    # Size of kernel
    x_kernel_width,
    y_kernel_width,
    z_kernel_width,
    # Size of blocks to load
    X_BLOCK_WIDTH: tl.constexpr,
    Y_BLOCK_WIDTH: tl.constexpr,
    Z_BLOCK_WIDTH: tl.constexpr,
    pts_per_grid,  # Determined via heuristic
    beta=1.0,  # For kernel=kaiser_bessel
):
    """ """
    size = x_size * y_size * z_size

    pid_0 = tl.program_id(0)
    grids_per_batch = tl.cast(tl.ceil(npts / pts_per_grid), tl.int32)
    N, grid_start = pid_0 // grids_per_batch, pid_0 % grids_per_batch
    pts_lower, pts_upper = pts_per_grid * grid_start, pts_per_grid * (grid_start + 1)

    x_base_range = tl.arange(0, X_BLOCK_WIDTH)
    y_base_range = tl.arange(0, Y_BLOCK_WIDTH)
    z_base_range = tl.arange(0, Z_BLOCK_WIDTH)
    if is_complex:
        # Last dimension has double size because (real, imag) are interleaved
        # Overall size doubles as a result
        size = 2 * size

    in_batch_offset = N * size
    out_batch_offset = N * npts

    for p in range(pts_lower, pts_upper):
        if p < npts:
            # Load target point
            x_target = tl.load(pts_ptr + 3 * p)
            y_target = tl.load(pts_ptr + 3 * p + 1)
            z_target = tl.load(pts_ptr + 3 * p + 2)

            weights, x_range, y_range, z_range, x_mask, y_mask, z_mask = weights3d(
                x_target,
                y_target,
                z_target,
                x_kernel_width,
                y_kernel_width,
                z_kernel_width,
                x_base_range,
                y_base_range,
                z_base_range,
                KERNEL,
                NORM,
                beta,
            )

            if is_complex:
                x_range_cplx = x_range  # [width]
                y_range_cplx = y_range
                x_mask_cplx = x_mask
                y_mask_cplx = y_mask
                # Pytorch interleaved indexing
                # Only applies to last dimension
                z_range_real = 2 * z_range  # 2 is for real/complex, not dimension
                z_range_imag = 2 * z_range + 1
                z_range_cplx = tl.join(z_range_real, z_range_imag)  # [width, 2]
                z_mask_cplx = tl.join(z_mask, z_mask)
                if PAD_MODE == "zero":
                    x_mask_cplx &= (x_range_cplx >= 0) & (x_range_cplx < x_size)
                    y_mask_cplx &= (y_range_cplx >= 0) & (y_range_cplx < y_size)
                    z_mask_cplx &= (z_range_cplx >= 0) & (z_range_cplx < (2 * z_size))
                elif PAD_MODE == "circular":
                    x_range_cplx = mod_pos(x_range_cplx, x_size)
                    y_range_cplx = mod_pos(y_range_cplx, y_size)
                    z_range_cplx = mod_pos(z_range_cplx, 2 * z_size)

                grid_range_cplx = (
                    x_range_cplx[:, None, None, None] * y_size
                    + y_range_cplx[None, :, None, None]
                ) * z_size * 2 + z_range_cplx[None, None, :, :]
                grid_mask_cplx = (
                    x_mask_cplx[:, None, None, None] & y_mask_cplx[None, :, None, None]
                ) & z_mask_cplx[None, None, :, :]

                # Load
                grid_cplx = tl.load(
                    in_ptr + in_batch_offset + grid_range_cplx, grid_mask_cplx
                )
                # Split and process separately
                grid_real, grid_imag = tl.split(grid_cplx)
                mask_real, mask_imag = tl.split(grid_mask_cplx)
                out_real = tl.sum(weights * grid_real * mask_real)
                out_imag = tl.sum(weights * grid_imag * mask_imag)
                tl.store(out_ptr + 2 * (out_batch_offset + p), out_real)
                tl.store(out_ptr + 2 * (out_batch_offset + p) + 1, out_imag)

            else:
                # Normal indexing
                if PAD_MODE == "zero":
                    x_mask &= (x_range >= 0) & (x_range < x_size)
                    y_mask &= (y_range >= 0) & (y_range < y_size)
                    z_mask &= (z_range >= 0) & (z_range < z_size)
                elif PAD_MODE == "circular":
                    x_range = mod_pos(x_range, x_size)
                    y_range = mod_pos(y_range, y_size)
                    z_range = mod_pos(z_range, z_size)

                grid_range = (
                    x_range[:, None, None] * y_size + y_range[None, :, None]
                ) * z_size + z_range[None, None, :]
                grid_mask = (
                    x_mask[:, None, None]
                    & y_mask[None, :, None]
                    & z_mask[None, None, :]
                )

                # Load
                grid = tl.load(in_ptr + in_batch_offset + grid_range, grid_mask)

                # Process jointly
                out = tl.sum(weights * grid)
                tl.store(out_ptr + out_batch_offset + p, out)


# @triton.heuristics(
#     values={
#         "pts_per_grid": lambda args: max(
#             1, triton.cdiv(args["npts"] * args["nbatch"], TRITON_MAX_GRID_SIZE)
#         ),
#     },
# )
# @triton.jit
# def _ungridnd(
#     in_ptr,
#     pts_ptr,
#     out_ptr,
#     size_ptr,
#     kernel_width_ptr,
#     nbatch,
#     npts,
#     KERNEL,
#     NORM,
#     PAD_MODE,
#     is_complex,  # bool
#     BLOCK_WIDTH,
#     pts_per_grid,  # Determined via heuristic
#     beta,  # For kernel=kaiser_bessel
# ):
#     """
#     Recursive version
#     """
#     ...


# @triton.jit
# def ungrid_nd_helper(
#     in_ptr, kernel_width_ptr, size_ptr, dim, BLOCK_WIDTH, KERNEL, beta
# ):
#     kernel_width = tl.load(kernel_width_ptr)
#     size = tl.load(size_ptr)
#     vals = tl.zeros(BLOCK_WIDTH)
#     if dim == 1:
#         load_range = tl.arange(0, BLOCK_WIDTH)
#         vals = tl.load(in_ptr + load_range)
#     else:
#         for i in range(kernel_width):
#             val = ungrid_nd_helper(
#                 in_ptr + i * size,
#                 kernel_width_ptr + 1,
#                 size_ptr + 1,
#                 dim - 1,
#                 KERNEL,
#                 beta,
#             )
#             # Jank version of "append"
#             vals += val * one_hot(i, BLOCK_WIDTH)
#
#     # Run normal 1d ungrid
#     target = tl.load(pts_ptr + p)
#     weights, x_range, x_mask = weights1d(target, kernel_width, base_range, KERNEL, beta)
#     # TODO: return dot product of vals and weights


@triton.jit
def one_hot(i, BLOCK_WIDTH: tl.constexpr, dtype: tl.constexpr = tl.float32):
    idx = tl.arange(0, BLOCK_WIDTH)
    return tl.where(idx == i, 1, 0).to(dtype)


if TRITON_ENABLED:
    UNGRID = {1: _ungrid1d, 2: _ungrid2d, 3: _ungrid3d}
else:
    UNGRID = {}


def prep_ungrid_shapes(vals, locs, width):
    ndim = locs.shape[-1]
    grid_size = tuple(vals.shape[-ndim:])

    # Flatten input vals
    batch_shape = vals.shape[:-ndim]
    if ndim < len(vals.shape):
        vals_flat = vals.flatten(0, len(vals.shape) - ndim - 1)
    else:
        vals_flat = vals[None]
    nbatch = vals_flat.shape[0]

    # Ensure locs are in [0, grid_size-1] in each dimension
    locs = torch.remainder(locs, torch.tensor(grid_size, device=locs.device))

    # Handle locs shapes
    locs_batch_shape = locs.shape[:-1]
    npts = prod(locs_batch_shape)

    # Complex input
    is_complex = torch.is_complex(vals)

    # Ensure kernel width is a tuple
    if isinstance(width, float):
        width = (width,) * ndim
    elif isinstance(width, tuple) and len(width) != ndim:
        raise ValueError(
            f"If width specified as tuple it must be same length as grid size but got len(width) = {len(width)} and len(grid_size) = {ndim}"
        )

    return (
        vals_flat,
        locs,
        {
            "ndim": ndim,
            "grid_size": grid_size,
            "width": width,
            "nbatch": nbatch,
            "batch_shape": batch_shape,
            "is_complex": is_complex,
            "locs_batch_shape": locs_batch_shape,
            "npts": npts,
        },
    )


def ungrid_torch(
    vals: Inexact[Tensor, "B ..."],
    locs: Float[Tensor, "... D"],
    grid_size: tuple[int, ...],
    width: tuple[float, ...],
    kernel: str = "kaiser_bessel",
    norm: str = "1",
    pad_mode: Literal["zero", "circular"] = "circular",
    batch_size: int = 2**20,
    **kernel_params,
):
    """Torch fallback

    Eventually, may want to use triton's CPU backend

    pad_mode : 'zero' or 'circular'
        Type of edge behavior to use
    batch size : int
        number of points to compute over at once
    """
    kernel_fn = get_kernel_fn(kernel, kernel_params)

    # Define helper vars
    nbatch = vals.shape[0]
    ndim = locs.shape[-1]
    locs_batch_shape = locs.shape[:-1]
    npts = prod(locs_batch_shape)
    device = vals.device
    grid_size = torch.tensor(grid_size, device=vals.device)

    # Get difference grid
    diff = torch.meshgrid(*(torch.arange(w + 1) for w in width), indexing="ij")
    # patch_shape = diff[0].shape
    # [prod(patch_shape), ndim]
    diff = torch.stack(diff, axis=-1).reshape(-1, ndim).to(device)
    radius = torch.tensor(width, device=device) / 2
    # [locs_batch_shape, ndim]
    locs_lower = torch.ceil(locs - radius).to(torch.int64)

    # For loop for memory purposes
    # Flatten locs batch shape
    locs = locs.reshape(-1, ndim)
    locs_lower = locs_lower.reshape(-1, ndim)
    # Create output
    out = torch.zeros(nbatch, npts, device=vals.device, dtype=vals.dtype)
    for p0, p1 in batch_iterator(npts, batch_size):
        grid_locs = locs_lower[p0:p1, None] + diff  # [prod(patch_shape), ndim]
        # Normalized delta(locations)
        weights, grid_locs, mask = weights_torch(
            locs[p0:p1],
            grid_locs,
            radius,
            norm,
            kernel_fn,
            grid_size,
            pad_mode,
        )
        grid_locs = (slice(None), *tuple(grid_locs[..., i] for i in range(ndim)))
        out[:, p0:p1] = torch.sum(weights * vals[grid_locs] * mask, dim=-1)
    out = out.reshape(nbatch, *locs_batch_shape)
    return out



from typing import Optional
from torch import Tensor

from copy import copy
from warnings import warn

from einops import repeat
import torch
import torch.nn as nn

from .nameddim import NS, Shape, ANY
from .namedlinop import NamedLinop, ND

__all__ = ["Diagonal"]


class Diagonal(NamedLinop):
    def __init__(
        self,
        weight: torch.Tensor,
        ioshape: Optional[Shape] = None,
        broadcast_dims: Optional[Shape] = None,
    ):
        if ioshape is not None and len(weight.shape) > len(ioshape):
            raise ValueError(
                f"All dimensions must be named or broadcastable, but got weight shape {weight.shape} and ioshape {ioshape}"
            )
        # if broadcast_dims is not None:
        #     warn(
        #         f"broadcast_dims argument is deprecated for torchlinops Diagonal but got {broadcast_dims}",
        #         DeprecationWarning,
        #         stacklevel=2,
        #     )
        super().__init__(NS(ioshape))
        self.weight = nn.Parameter(weight, requires_grad=False)
        # assert (
        #     len(self.ishape) >= len(self.weight.shape)
        # ), f"Weight cannot have fewer dimensions than the input shape: ishape: {self.ishape}, weight: {weight.shape}"
        broadcast_dims = broadcast_dims if broadcast_dims is not None else []
        if ANY in self.ishape:
            broadcast_dims.append(ANY)
        self._shape.add("broadcast_dims", broadcast_dims)

    @classmethod
    def from_weight(
        cls,
        weight: Tensor,
        weight_shape: Shape,
        ioshape: Shape,
        shape_kwargs: Optional[dict] = None,
    ):
        shape_kwargs = shape_kwargs if shape_kwargs is not None else {}
        if len(weight.shape) > len(ioshape):
            raise ValueError(
                f"All dimensions must be named or broadcastable, but got weight shape {weight.shape} and ioshape {ioshape}"
            )
        weight = repeat(
            weight,
            f"{' '.join(weight_shape)} -> {' '.join(ioshape)}",
            **shape_kwargs,
        )
        return cls(weight, ioshape)

    @property
    def broadcast_dims(self):
        return self._shape.lookup("broadcast_dims")

    @broadcast_dims.setter
    def broadcast_dims(self, val):
        self._shape.broadcast_dims = val

    # Override shape setters too
    @NamedLinop.ishape.setter
    def ishape(self, val):
        self._shape.ishape = val
        self._shape.oshape = val

    @NamedLinop.oshape.setter
    def oshape(self, val):
        self._shape.oshape = val
        self._shape.ishape = val

    def forward(self, x):
        return self.fn(self, x, self.weight)

    @staticmethod
    def fn(linop, x, /, weight):
        return x * weight

    @staticmethod
    def adj_fn(linop, x, /, weight):
        return x * torch.conj(weight)

    @staticmethod
    def normal_fn(linop, x, /, weight):
        return x * torch.abs(weight) ** 2

    def adjoint(self):
        adj = copy(self)
        adj.weight = nn.Parameter(
            self.weight.conj(),
            requires_grad=self.weight.requires_grad,
        )
        return adj

    def normal(self, inner=None):
        if inner is None:
            normal = copy(self)
            normal.weight = nn.Parameter(
                torch.abs(self.weight) ** 2,
                requires_grad=self.weight.requires_grad,
            )
            return normal
        return super().normal(inner)

    def split_forward(self, ibatch, obatch):
        weight = self.split_forward_fn(ibatch, obatch, self.weight)
        split = copy(self)
        split.weight = nn.Parameter(weight, requires_grad=self.weight.requires_grad)
        return split

    def split_forward_fn(self, ibatch, obatch, /, weight):
        assert ibatch == obatch, "Diagonal linop must be split identically"
        # Filter out broadcastable dims
        ibatch = [
            slice(None) if dim in self.broadcast_dims else slc
            for slc, dim in zip(ibatch, self.ishape)
        ]
        return weight[ibatch[-len(weight.shape) :]]

    def size(self, dim: str):
        return self.size_fn(dim, self.weight)

    def size_fn(self, dim: str, weight):
        if dim in self.ishape:
            n_broadcast = len(self.ishape) - len(weight.shape)
            if self.ishape.index(dim) < n_broadcast or dim in self.broadcast_dims:
                return None
            else:
                return weight.shape[self.ishape.index(dim) - n_broadcast]
        return None

    def __pow__(self, exponent):
        new = copy(self)
        new.weight = nn.Parameter(
            self.weight**exponent,
            requires_grad=self.weight.requires_grad,
        )
        return new



from typing import Any
from types import SimpleNamespace
from functools import wraps

__all__ = ["fake_triton", "fake_tl"]


# Replace all triton decorators with not implemented
def not_implemented_wrapper(*args, **kwargs):
    raise NotImplementedError()


def not_implemented_decorator(f):
    return wraps(f)(not_implemented_wrapper)


fake_triton = SimpleNamespace()
fake_triton.jit = lambda f: wraps(f)(not_implemented_wrapper)
fake_triton.heuristics = lambda *args, **kwargs: not_implemented_decorator

fake_tl = SimpleNamespace()
fake_tl.constexpr = Any
fake_tl.float32 = float



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from pathlib import Path
import sys

sys.path.insert(0, str(Path(__file__).parent.parent.parent / "src"))

import torchlinops

project = "torch-named-linops"
copyright = "2024, Mark Nishimura"
author = "Mark Nishimura"
version = torchlinops.__version__
release = torchlinops.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    # "numpydoc",
    # "myst_parser",  # Allow markdown, comment out if myst_nb is enabled
    "myst_nb",  # Execute code blocks in docs
]

templates_path = ["_templates"]
exclude_patterns = []

myst_enable_extensions = [
    "colon_fence",
]

# Options for autodoc
# Prevent inheriting from torch base modules
autodoc_inherit_docstrings = False
autodoc_default_options = {
    "members": False,
    "undoc-members": False,
    "private-members": False,
    "inherited-members": False,
    "show-inheritance": False,
}
autosummary_generate = True
templates_path = ["_templates"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ["_static"]
html_theme = "furo"
html_title = "torch-named-linops"
html_theme_options = {
    "sidebar_hide_name": False,
}
html_css_files = ["codecells.css"]



"""Differentiable Block/Unblock functions"""

from typing import Optional, Literal
from jaxtyping import Bool
from torch import Tensor
from torch.autograd import Function

import torch
from .unfold import unfold
from .fold import fold

__all__ = ["array_to_blocks", "blocks_to_array"]


class ArrayToBlocksFn(Function):
    """Equal to block/unfold"""

    @staticmethod
    def forward(
        input: Tensor,
        block_shape,  # tuple
        stride,  # tuple
        mask,  # Bool[Tensor]
        out,  # Optional[Tensor]
    ) -> Tensor:
        """Remembered something about Function not liking type annotations?"""
        output = unfold(input, block_shape, stride, mask, out)
        output.requires_grad_(input.requires_grad)
        return output

    @staticmethod
    def setup_context(ctx, inputs, output):
        # Unpack input and output
        input, block_shape, stride, mask, out = inputs

        # Save for backward pass
        ctx.im_size = tuple(input.shape[-len(block_shape) :])
        ctx.block_shape = block_shape
        ctx.stride = stride
        ctx.save_for_backward(mask)

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = grad_block_shape = grad_stride = grad_mask = grad_out = None

        if ctx.needs_input_grad[0]:
            mask = ctx.saved_tensors[0]
            grad_input = fold(
                grad_output,
                ctx.im_size,
                ctx.block_shape,
                ctx.stride,
                mask,
            )
        return grad_input, grad_block_shape, grad_stride, grad_mask, grad_out


def array_to_blocks(
    input,
    block_shape: tuple[int, ...],
    stride: Optional[tuple[int, ...]] = None,
    mask: Optional[Bool[Tensor, "..."]] = None,
    out: Optional[Tensor] = None,
):
    """Wrapper for default arguments"""
    return ArrayToBlocksFn.apply(input, block_shape, stride, mask, out)


class BlocksToArrayFn(Function):
    """Equal to block_adjoint/fold"""

    @staticmethod
    def forward(
        input: Tensor,
        im_size,  #  tuple
        block_shape,  # tuple
        stride,  # tuple
        mask,  # Tensor
        out,  # Optional[Tensor]
    ) -> Tensor:
        output = fold(input, im_size, block_shape, stride, mask, out)
        output.requires_grad_(input.requires_grad)
        return output

    @staticmethod
    def setup_context(ctx, inputs, output):
        input, im_size, block_shape, stride, mask, out = inputs

        # Save for backward pass
        ctx.block_shape = block_shape
        ctx.stride = stride
        ctx.save_for_backward(mask)

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = grad_im_size = grad_block_shape = grad_stride = grad_mask = (
            grad_out
        ) = None
        if ctx.needs_input_grad[0]:
            mask = ctx.saved_tensors[0]
            grad_input = unfold(grad_output, ctx.block_shape, ctx.stride, mask)
        return (
            grad_input,
            grad_im_size,
            grad_block_shape,
            grad_stride,
            grad_mask,
            grad_out,
        )


def blocks_to_array(
    input,
    im_size: tuple,
    block_shape: tuple,
    stride: Optional[tuple] = None,
    mask: Optional[Bool[Tensor, "..."]] = None,
    out: Optional[Tensor] = None,
):
    """Wrapper for default arguments"""
    return BlocksToArrayFn.apply(input, im_size, block_shape, stride, mask, out)


def get_norm_weights(
    im_size: tuple[int, ...],
    block_shape: tuple[int, ...],
    stride: Optional[tuple[int, ...]] = None,
    mask: Optional[Bool[Tensor, "..."]] = None,
    device: torch.device = "cpu",
):
    """Compute normalizing weights

    1./weights * blocks_to_array(array_to_blocks(x)) == x

    """
    x = torch.ones(im_size, device=device)
    Ax = array_to_blocks(x, block_shape, stride, mask)
    weights = blocks_to_array(Ax, im_size, block_shape, stride, mask)
    return weights



from typing import Optional
from copy import copy
import torch
from .namedlinop import NamedLinop
from .identity import Identity
from .nameddim import ELLIPSES, NS, Shape
from torchlinops.utils import INDENT

__all__ = ["ToDevice"]


class ToDevice(NamedLinop):
    def __init__(
        self,
        idevice: torch.device | str,
        odevice: torch.device | str,
        ioshape: Optional[Shape] = None,
    ):
        super().__init__(NS(ioshape))
        self.idevice = torch.device(idevice)
        self.odevice = torch.device(odevice)

    @staticmethod
    def fn(linop, x, /):
        if x.device != linop.idevice:
            raise RuntimeError(
                f"Got input to ToDevice on {x.device} but expected {linop.idevice}"
            )
        return x.to(linop.odevice)

    @staticmethod
    def adj_fn(linop, x, /):
        if x.device != linop.odevice:
            raise RuntimeError(
                f"Got input to ToDevice on {x.device} but expected {linop.odevice}"
            )
        return x.to(linop.idevice)

    def adjoint(self):
        adj = copy(self)
        adj._shape = adj._shape.H
        adj.idevice, adj.odevice = self.odevice, self.idevice
        return adj

    def normal(self, inner=None):
        if inner is None:
            return Identity()
        return super().normal(inner)

    def split_forward(self, ibatch, obatch):
        """Return a new instance"""
        return copy(self)

    def __repr__(self):
        """Helps prevent recursion error caused by .H and .N"""
        out = f"({self.idevice} -> {self.odevice})"
        out = INDENT.indent(out)
        return out



from typing import Optional, Literal

from pathlib import Path
import copy
import gc
import logging

import torch

import cupy as cp
import numpy as np
from easydict import EasyDict

from ._defaults import default_to
from ._log import Indenter

logger = logging.getLogger("torchmri.utils")

__all__ = ["benchmark", "benchmark_and_summarize"]


def benchmark_and_summarize(
    fn,
    *args,
    num_iters: int = 10,
    ignore_first: int = 0,
    backend: Literal["torch", "cupy"] = "torch",
    name: str = None,
    **kwargs,
):
    """Convenience function"""
    result, output = benchmark(
        fn, *args, num_iters=num_iters, backend=backend, **kwargs
    )
    if name is None:
        name = fn.__name__
    summarize(result, name=name, ignore_first=ignore_first)
    return result, output


def benchmark(
    fn,
    *args,
    num_iters: int = 10,
    backend: Literal["torch", "cupy"] = "torch",
    **kwargs,
):
    """Benchmark a function called with some arguments.

    Defaults to torch benchmarking
    """
    if backend == "torch":
        backend = TorchHandler()
    elif backend == "cupy":
        backend = CupyHandler()
    else:
        raise ValueError(f"Unrecognized backend type {backend}")
    fn_result = fn(*args, **kwargs)  # Warmup
    backend.bench_start()
    for _ in range(num_iters):
        backend.trial_start()
        fn(*args, **kwargs)
        backend.trial_end()
    backend.bench_end()
    return backend.result, fn_result


def summarize(benchmark_result, name: str, ignore_first: int = 0):
    """Summarize the results from benchmark"""
    with Indenter() as indent:
        print(name)
        with indent:
            timings_ms = benchmark_result["timings_ms"][ignore_first:]
            indent.print(f"Mean Time: {np.mean(timings_ms):0.3f} ms")
            indent.print(f"Min Time: {np.min(timings_ms):0.3f} ms")
            indent.print(f"Max Time: {np.max(timings_ms):0.3f} ms")
            indent.print(f"Memory: {benchmark_result['max_mem_bytes']} bytes")


class TorchHandler:
    def __init__(self, memory_snapshot_file: Optional[Path] = "memory.pkl"):
        self.reset()
        self.memory_snapshot_file = memory_snapshot_file

    def reset(self):
        self._start_event = None
        self._end_event = None

        self.result = EasyDict(
            {
                "timings_ms": [],
                "max_mem_bytes": None,
            }
        )

    def bench_start(self, *args, **kwargs):
        self.reset()
        gc.disable()
        torch.cuda.reset_peak_memory_stats()
        torch.cuda.memory._record_memory_history(
            max_entries=100000,
        )

    def bench_end(self, *args, **kwargs):
        self.result.max_mem_bytes = torch.cuda.max_memory_allocated()
        gc.enable()
        logger.info(f"Max memory allocated: {self.result.max_mem_bytes}")
        try:
            torch.cuda.memory._dump_snapshot(f"{str(self.memory_snapshot_file)}")
        except Exception as e:
            logger.error(f"Failed to capture memory snapshot {e}")

        # Stop recording memory snapshot history.
        torch.cuda.memory._record_memory_history(enabled=None)

    def trial_start(self, event=None, i=None):
        self._start_event = torch.cuda.Event(enable_timing=True)
        self._end_event = torch.cuda.Event(enable_timing=True)
        self._start_event.record()

    def trial_end(self, event=None, i=None):
        self._end_event.record()
        torch.cuda.synchronize()
        time = self._start_event.elapsed_time(self._end_event)
        logger.debug(f"{event}: {time}")
        self.result.timings_ms.append(time)

    def collect_results(self, event, data):
        return {"torch": copy.deepcopy(self.result)}


class CupyHandler:
    """Benchmarking class for CuPy-based functions

    Usage:


    """

    def __init__(self):
        self.reset()

    def reset(self):
        self._start_event = None
        self._end_event = None
        self._mempool = None

        self.result = EasyDict(
            {
                "timings_ms": [],
                "max_mem_bytes": None,
            }
        )

    def bench_start(self, *args, **kwargs):
        self.reset()
        gc.disable()
        self._mempool = cp.get_default_memory_pool()
        self._mempool.free_all_blocks()

    def bench_end(self, *args, **kwargs):
        self.result.max_mem_bytes = self._mempool.total_bytes()
        gc.enable()

    def trial_start(self, event=None, i=None):
        self._start_event = cp.cuda.Event(disable_timing=False)
        self._end_event = cp.cuda.Event(disable_timing=False)
        self._start_event.record()

    def trial_end(self, event=None, i=None):
        self._end_event.record()
        self._end_event.synchronize()
        time = cp.cuda.get_elapsed_time(self._start_event, self._end_event)  # ms
        self.result.timings_ms.append(time)

    def collect_results(self, *args, **kwargs):
        return {"cupy": copy.deepcopy(self.result)}



from typing import Optional

import torch
import torch.nn as nn

from torchlinops.utils import default_to

from .diagonal import Diagonal
from .nameddim import NDorStr, Shape


__all__ = ["Scalar"]


class Scalar(Diagonal):
    """The result of scalar multiplication

    A Diagonal linop that is trivially splittable.
    """

    def __init__(self, weight, ioshape: Optional[Shape] = None):
        if not isinstance(weight, torch.Tensor):
            weight = torch.tensor(weight)
        ioshape = default_to(("...",), ioshape)
        super().__init__(weight, ioshape=ioshape)

    def split_forward_fn(self, ibatch, obatch, /, weight):
        assert ibatch == obatch, "Scalar linop must be split identically"
        return weight

    def size_fn(self, dim: str, weight):
        return None



import pytest


@pytest.fixture
def small1d():
    N = (2,)
    grid_size = (10,)
    # npts = 20
    locs_batch_size = (20,)
    width = 4.0

    spec = {
        "N": N,
        "grid_size": grid_size,
        "locs_batch_size": locs_batch_size,
        "width": width,
    }
    return spec


@pytest.fixture
def medium1d():
    N = (2, 2)
    grid_size = (24,)
    locs_batch_size = (6, 11)
    width = 4.0

    spec = {
        "N": N,
        "grid_size": grid_size,
        "locs_batch_size": locs_batch_size,
        "width": width,
    }
    return spec


@pytest.fixture
def large1d():
    N = (2,)
    grid_size = (220,)
    locs_batch_size = (8, 8, 4)
    width = 4.0

    spec = {
        "N": N,
        "grid_size": grid_size,
        "locs_batch_size": locs_batch_size,
        "width": width,
    }
    return spec


@pytest.fixture
def tiny2d():
    N = (2,)
    grid_size = (2, 2)
    locs_batch_size = (1,)
    width = 2.0

    spec = {
        "N": N,
        "grid_size": grid_size,
        "locs_batch_size": locs_batch_size,
        "width": width,
    }
    return spec


@pytest.fixture
def small2d():
    N = (2, 2)
    grid_size = (10, 8)
    locs_batch_size = (9,)
    width = 4.0

    spec = {
        "N": N,
        "grid_size": grid_size,
        "locs_batch_size": locs_batch_size,
        "width": width,
    }
    return spec


@pytest.fixture
def medium2d():
    N = (2, 2)
    grid_size = (124, 125)
    locs_batch_size = (3, 19)
    # Kernel size larger than 4 might cause test to fail
    # because of numerical issues
    width = 4.0

    spec = {
        "N": N,
        "grid_size": grid_size,
        "locs_batch_size": locs_batch_size,
        "width": width,
    }
    return spec


@pytest.fixture
def large2d():
    N = (2, 2)
    grid_size = (220, 220)
    locs_batch_size = (4, 3, 10)
    width = 4.0

    spec = {
        "N": N,
        "grid_size": grid_size,
        "locs_batch_size": locs_batch_size,
        "width": width,
    }
    return spec


@pytest.fixture
def small3d():
    N = (1, 1)
    grid_size = (10, 8, 9)
    locs_batch_size = (3, 5)
    width = 4.0

    spec = {
        "N": N,
        "grid_size": grid_size,
        "locs_batch_size": locs_batch_size,
        "width": width,
    }
    return spec


@pytest.fixture
def medium3d():
    N = (2, 2)
    grid_size = (33, 35, 37)
    locs_batch_size = (3, 19)
    width = 4.0

    spec = {
        "N": N,
        "grid_size": grid_size,
        "locs_batch_size": locs_batch_size,
        "width": width,
    }
    return spec


@pytest.fixture
def large3d():
    N = (1, 1)
    grid_size = (220, 220, 220)
    locs_batch_size = (48, 500, 1600)
    width = 4.0

    spec = {
        "N": N,
        "grid_size": grid_size,
        "locs_batch_size": locs_batch_size,
        "width": width,
    }
    return spec



import pytest

from math import prod

import torch
import sigpy as sp

from torchlinops.functional import fold
from torchlinops.functional._unfold.nblocks import get_nblocks
from torchlinops.utils import from_pytorch, to_pytorch


PYTEST_GPU_MARKS = [
    pytest.mark.gpu,
    pytest.mark.skipif(
        not torch.cuda.is_available(), reason="GPU is required but not available"
    ),
]


@pytest.mark.parametrize("dev", ["cpu", pytest.param("cuda", marks=PYTEST_GPU_MARKS)])
@pytest.mark.parametrize("dtype", ["float32", "float64", "complex64"])
@pytest.mark.parametrize(
    "spec",
    [
        "small1d",
        "medium1d",
        pytest.param("large1d", marks=PYTEST_GPU_MARKS),
        "tiny2d",
        "small2d",
        "medium2d",
        pytest.param("large2d", marks=PYTEST_GPU_MARKS),
        pytest.param("small3d", marks=PYTEST_GPU_MARKS),
        pytest.param("medium3d", marks=PYTEST_GPU_MARKS),
        pytest.param("large3d", marks=PYTEST_GPU_MARKS + [pytest.mark.slow]),
        pytest.param("verylarge3d", marks=PYTEST_GPU_MARKS + [pytest.mark.slow]),
    ],
)
def test_fold(dev, dtype, spec, request):
    spec = request.getfixturevalue(spec)
    device = torch.device(dev)
    if dtype == "float32":
        dtype = torch.float32
    elif dtype == "float64":
        dtype = torch.float64
    elif dtype == "complex64":
        dtype = torch.complex64

    spec["nblocks"] = get_nblocks(spec["shape"], spec["block_size"], spec["stride"])

    ishape = (*spec["N"], *spec["nblocks"], *spec["block_size"])
    oshape = (*spec["N"], *spec["shape"])

    x = torch.arange(prod(ishape)).reshape(ishape)
    # x = torch.ones(prod(ishape)).reshape(ishape)

    x = x.to(device).to(dtype)

    y_th = fold(x, spec["shape"], spec["block_size"], spec["stride"])

    x = from_pytorch(x)
    y_sp = sp.blocks_to_array(x, oshape, spec["block_size"], spec["stride"])
    assert torch.allclose(y_th, to_pytorch(y_sp))



import pytest

import torch

from torchlinops import Diagonal, NUFFT, Concat
from torchlinops.functional._interp.tests._valid_pts import get_valid_locs


def test_sense_mri():
    # Define shapes
    grid_size = (128, 128, 128)
    Nx, Ny, Nz = grid_size
    R, K = (6, 28)
    C = 11

    ishape = (Nx, Ny, Nz)
    oshape = (C, R, K)
    x = torch.randn(*ishape)
    S_weight = torch.randn(C, Nx, Ny, Nz)
    S = Diagonal.from_weight(
        S_weight,
        weight_shape=("C", "Nx", "Ny", "Nz"),
        ioshape=("C", "Nx", "Ny", "Nz"),
    )
    locs = get_valid_locs(
        (R, K),
        grid_size,
        ndim=len(grid_size),
        width=4.0,
        device="cpu",
        centered=True,
    )
    F = NUFFT(locs, grid_size, output_shape=("R", "K"))

    A = F @ S
    b = A(x)


def test_subspace_mri(): ...


def test_subspace_timeseg_mri(): ...



import torch

__all__ = ["end_pad_with_zeros"]


def end_pad_with_zeros(t, dim, pad_length):
    """Extend a tensor with a specified number of zeros along a dimension"""
    pad_shape = list(t.shape)
    pad_shape[dim] = pad_length
    pad = torch.zeros(*tuple(pad_shape), dtype=t.dtype, device=t.device)
    t_pad = torch.concatenate((t, pad), dim=dim)
    return t_pad



"""Differentiable forms of Grid/Ungrid"""

from typing import Optional, Literal
from jaxtyping import Inexact, Float
from torch import Tensor
from torch.autograd import Function

import torch
from .ungrid import ungrid
from .grid import grid

__all__ = ["interpolate", "interpolate_adjoint"]


class InterpolateFn(Function):
    """Equal to block/unfold"""

    @staticmethod
    def forward(
        vals: Inexact[Tensor, "..."],
        locs: Float[Tensor, "... D"],
        width: float | tuple[float, ...],
        kernel: str,
        norm: str,
        pad_mode: str,
        kernel_params: dict,
    ) -> Tensor:
        """Remembered something about Function not liking type annotations?"""
        output = ungrid(vals, locs, width, kernel, norm, pad_mode, kernel_params)
        output.requires_grad_(vals.requires_grad)
        return output

    @staticmethod
    def setup_context(ctx, inputs, output):
        # Unpack input and output
        vals, locs, width, kernel, norm, pad_mode, kernel_params = inputs

        # Save for backward pass
        ndim = locs.shape[-1]
        ctx.grid_size = vals.shape[-ndim:]
        ctx.width = width
        ctx.kernel = kernel
        ctx.norm = norm
        ctx.pad_mode = pad_mode
        ctx.kernel_params = kernel_params
        ctx.save_for_backward(locs)

    @staticmethod
    def backward(ctx, grad_output):
        """"""
        grad_vals = grad_locs = grad_width = grad_kernel = grad_norm = grad_pad_mode = (
            grad_kernel_params
        ) = None
        if ctx.needs_input_grad[0]:
            locs = ctx.saved_tensors[0]
            grad_vals = grid(
                grad_output,
                locs,
                ctx.grid_size,
                ctx.width,
                ctx.kernel,
                ctx.norm,
                ctx.pad_mode,
                ctx.kernel_params,
            )
        return (
            grad_vals,
            grad_locs,
            grad_width,
            grad_kernel,
            grad_norm,
            grad_pad_mode,
            grad_kernel_params,
        )


def interpolate(
    vals: Inexact[Tensor, "..."],
    locs: Float[Tensor, "... D"],
    width: float | tuple[float, ...],
    kernel="kaiser_bessel",
    norm: str = "1",
    pad_mode: str = "circular",
    kernel_params: dict = None,
):
    """Wrapper for default arguments"""
    return InterpolateFn.apply(vals, locs, width, kernel, norm, pad_mode, kernel_params)


class InterpolateAdjointFn(Function):
    """Equal to block_adjoint/fold"""

    @staticmethod
    def forward(
        vals: Inexact[Tensor, "..."],
        locs: Float[Tensor, "... D"],
        grid_size: tuple[int, ...],
        width: float | tuple[float, ...],
        kernel: str,
        norm: str,
        pad_mode: str,
        kernel_params: dict,
    ) -> Tensor:
        output = grid(
            vals,
            locs,
            grid_size,
            width,
            kernel,
            norm,
            pad_mode,
            kernel_params,
        )
        output.requires_grad_(vals.requires_grad)
        return output

    @staticmethod
    def setup_context(ctx, inputs, output):
        vals, locs, grid_size, width, kernel, norm, pad_mode, kernel_params = inputs

        # Save for backward pass
        ctx.width = width
        ctx.kernel = kernel
        ctx.norm = norm
        ctx.pad_mode = pad_mode
        ctx.kernel_params = kernel_params
        ctx.save_for_backward(locs)

    @staticmethod
    def backward(ctx, grad_output):
        grad_vals = grad_locs = grad_grid_size = grad_width = grad_kernel = (
            grad_norm
        ) = grad_pad_mode = grad_kernel_params = None
        if ctx.needs_input_grad[0]:
            locs = ctx.saved_tensors[0]
            grad_vals = ungrid(
                grad_output,
                locs,
                ctx.width,
                ctx.kernel,
                ctx.norm,
                ctx.pad_mode,
                ctx.kernel_params,
            )
        return (
            grad_vals,
            grad_locs,
            grad_grid_size,
            grad_width,
            grad_kernel,
            grad_norm,
            grad_pad_mode,
            grad_kernel_params,
        )


def interpolate_adjoint(
    vals: Inexact[Tensor, "..."],
    locs: Float[Tensor, "... D"],
    grid_size: tuple[int, ...],
    width: float | tuple[float, ...],
    kernel: str = "kaiser_bessel",
    norm: str = "1",
    pad_mode: str = "circular",
    kernel_params: dict = None,
):
    """Wrapper for default arguments"""
    return InterpolateAdjointFn.apply(
        vals,
        locs,
        grid_size,
        width,
        kernel,
        norm,
        pad_mode,
        kernel_params,
    )



from warnings import warn
import torch
import torch.nn as nn
import torch.multiprocessing as mp
from torchlinops import Dense, Batch
from torch.profiler import profile, record_function, ProfilerActivity

try:
    mp.set_start_method("forkserver")
except RuntimeError:
    pass


def main():
    # Set up
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    B, M, N = 3, 10000, 1000
    weight = torch.randn(B, M, N, device=device)
    A = Dense(weight, ("B", "M", "N"), ishape=("N",), oshape=("B", "M"))
    A.to(device)
    A_batch = Batch(A, device, device, torch.float32, torch.float32, B=1)
    x = torch.randn(N, device=device)

    # Initialize pool
    pool = PersistentWorkerPool(A_batch._linops, do_profile=True)

    # Repeat to show warming up
    for i in range(10):
        inputs = [(x,) for _ in range(B)]
        # Issue a batch of data to the child processes
        for b in range(B):
            outputs = pool.parallel_run(inputs)
    pool.close()


class PersistentWorkerPool:
    def __init__(self, modules: list[nn.Module], do_profile: bool = False):
        self.workers = []
        for i, mod in enumerate(modules):
            self.workers.append(ModuleWorker(mod, rank=i, do_profile=do_profile))

    def __len__(self):
        return len(self.workers)

    def parallel_run(self, args_list: list[tuple]):
        """Run workers with arguments"""
        if len(args_list) != len(self):
            warn(
                f"Persistent Worker Pool called with different numbers of arguments {len(args_list)} than underlying workers: {len(self)}"
            )
        for args, worker in zip(args_list, self.workers):
            worker.call(args)

        outputs = []
        for worker in self.workers:
            outputs.append(worker.get())
        return outputs

    def close(self):
        for worker in self.workers:
            worker.stop()


# Special token for killing the worker

STOP = "STOP"


class ModuleWorker:
    """A torch module inside a process"""

    def __init__(self, mod: nn.Module, rank: int, do_profile):
        self.rank = rank
        self.do_profile = do_profile
        self.qin = mp.Queue()
        self.qout = mp.Queue()
        self.reset = mp.Event()
        self.all_done = mp.Event()

        # Initialize worker
        self._worker = mp.Process(
            target=_worker_main,
            args=(rank, self.qin, self.qout, self.reset, self.all_done, do_profile),
        )
        self._worker.start()
        self.qin.put(mod)

    def call(self, args: tuple):
        """Nonblocking"""
        self.qin.put(args)

    def get(self):
        """Blocking"""
        out = self.qout.get()
        self.reset.set()

    def stop(self):
        self.qin.put(STOP)
        self._worker.join()
        self._worker.close()  # For completeness


def _worker_main(
    rank: int,
    qin: mp.Queue,
    qout: mp.Queue,
    reset: mp.Event,
    all_done: mp.Event,
    do_profile: bool = False,
):
    mod = qin.get()
    if do_profile:
        with profile(
            activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
            on_trace_ready=lambda p: p.export_chrome_trace(
                f"./worker_{rank}_trace.json"
            ),
        ) as prof:
            while True:
                # Main loop
                # Ready state
                reset.clear()
                args = qin.get()
                if args == STOP:
                    break
                # Running state
                out = mod(*args)
                qout.put(out)
                # Resting state
                reset.wait()

    else:
        while True:
            # Main loop
            # Ready state
            args = qin.get()
            if args == STOP:
                break
            # Running state
            out = mod(*args)
            qout.put(out)
            # Resting state
            reset.wait()


if __name__ == "__main__":
    main()



from torch import Tensor

from functools import partial
import torch

from einops import rearrange
import torch.nn.functional as F
from torch.autograd.functional import vjp


def circular_pad(t: Tensor, padding: int | tuple):
    # Infer dimension
    if isinstance(padding, int):
        ndim = t.ndim
    else:  # isinstance(padding, tuple):
        if len(padding) % 2:
            raise ValueError(
                f"Padding must have even length but got {len(padding)}: {padding}"
            )
        ndim = len(padding) // 2

    # Flatten batch dim
    batch_shape = t.shape[:-ndim]
    if ndim < len(t.shape):
        t = t.flatten(0, len(t.shape) - ndim - 1)
    else:
        # Add fake batch dim
        t = t[None]

    # Flag for cycling the batch one more time at the end
    # Avoid uselessly cycling the batch if the initial ndim is <= 3
    cycle_batch = False
    while True:
        pad_dims = min(ndim, 3)
        pad = padding[: 2 * pad_dims]
        t = circular_pad_nd(t, pad, pad_dims)

        # Update
        ndim -= pad_dims
        if ndim == 0 and not cycle_batch:
            break

        padding = padding[2 * pad_dims :]

        # Cycle remaining dims to the back, padded dims to the front
        if pad_dims == 1:
            t = rearrange(t, "... x -> x ...")
        elif pad_dims == 2:
            t = rearrange(t, "... x y  -> x y ...")
        elif pad_dims == 3:
            t = rearrange(t, "... x y z -> x y z ...")
        cycle_batch = True
        if ndim == 0:
            break

    if cycle_batch:
        # Move batch dim back to front and unflatten
        # Batch dim should be the only one left
        t = rearrange(t, "... b -> b ...")
    t = t.reshape(*batch_shape, *t.shape[1:])
    return t


def circular_pad_nd(t: Tensor, padding: int | tuple, ndim: int):
    """
    Parameters
    ----------
    t : Tensor
        The tensor to be padded
    padding : int or tuple
        Follows same conventions as torch.nn.functional.pad
    """

    # Prep
    oshape = list(t.shape)
    for i, (padleft, padright) in enumerate(zip(padding[::2], padding[1::2])):
        oshape[-(i + 1)] += padleft + padright

    if ndim == 1:
        t = rearrange(t, "... x -> (...) x")
    elif ndim == 2:
        t = rearrange(t, "... x y -> (...) x y")
    elif ndim == 3:
        t = rearrange(t, "... x y z -> (...) x y z ")
    elif ndim > 3:
        raise NotImplementedError(
            "Circular padding not yet implemented for >3 dimensions"
        )
    else:
        raise ValueError(f"ndim must be positive int but got {ndim}")

    # Do pad
    t = F.pad(t, padding, mode="circular")

    # Postproc
    t = t.reshape(*oshape)
    return t


def circular_pad_adjoint(t: Tensor, padding: int | tuple):
    """Adjoint of circular pad
    Uses vjp for compactness of implementation (may not be the most performant)
    May also not play well with autograd (hopefully this doesn't matter)
    """
    # Prep
    ishape = list(t.shape)
    for i, (padleft, padright) in enumerate(zip(padding[::2], padding[1::2])):
        ishape[-(i + 1)] -= padleft + padright

    input_ = torch.zeros(*ishape, dtype=t.dtype, device=t.device)
    f = partial(circular_pad, padding=padding)
    t = vjp(f, inputs=input_, v=t, create_graph=True, strict=False)
    return t[1]



from typing import Literal
import pytest

from abc import ABC, abstractmethod

import torch
from torchlinops.utils import inner


class BaseNamedLinopTests(ABC):
    """Abstract base class for all tests that any linop should satisfy.

    To test a new linop:
    1. Create a class that inherits from this one
    2. Override the `linop_input_output` method to return a parameterized linop, and a test input and output`.
    3. If desired, add additional methods to the class (beginning with `test_`) to test additional functionality.
    """

    equality_check: Literal["exact", "approx"] = "exact"
    isclose_kwargs: dict = {}

    @pytest.fixture
    @abstractmethod
    def linop_input_output(self):
        """Create and return:
        1. A linop to test
        2. A tensor with the same shape as the linop's inputs
        3. A tensor with the same shape as the linop's outputs
        Ideally these are randomized in some fashion
        """
        pass

    def test_adjoint(self, linop_input_output):
        A, x, y = linop_input_output
        yAx = inner(y, A(x))
        xAHy = inner(A.H(y), x)
        if self.equality_check == "exact":
            assert yAx == xAHy
        elif self.equality_check == "approx":
            assert torch.isclose(yAx, xAHy, **self.isclose_kwargs).all()
        else:
            raise ValueError(f"Unrecognized equality_check mode: {self.equality_check}")

    def test_normal(self, linop_input_output):
        A, x, _ = linop_input_output
        AHAx = A.H(A(x))
        ANx = A.N(x)
        if self.equality_check == "exact":
            assert AHAx == ANx
        elif self.equality_check == "approx":
            assert torch.isclose(AHAx, ANx, **self.isclose_kwargs).all()
        else:
            raise ValueError(f"Unrecognized equality_check mode: {self.equality_check}")

    def test_adjoint_level2(self, linop_input_output):
        A, x, y = linop_input_output
        self.test_adjoint((A.H, y, x))
        self.test_normal((A.H, y, y))

    def test_normal_level2(self, linop_input_output):
        A, x, y = linop_input_output
        self.test_adjoint((A.N, x, x))
        self.test_normal((A.N, x, x))

    def test_split(self, linop_input_output):
        A, x, y = linop_input_output
        for dim in set(A.ishape + A.oshape):
            ...

    def test_backprop(self, linop_input_output):
        A, x, y = linop_input_output

        x = x.clone().requires_grad_(True)
        out = A.apply(x)
        if torch.is_complex(out):
            out.real.sum().backward()
        else:
            out.sum().backward()
        grad_out = torch.ones_like(out)
        assert x.grad.allclose(A.H.apply(grad_out), rtol=1e-3)



import torchlinops

# Global config variables
reduce_identity_in_normal = True


def inner_not_relevant(inner):
    return (inner is None) or (
        isinstance(inner, torchlinops.Identity) and reduce_identity_in_normal
    )



from typing import Optional
from torch import Tensor
from copy import copy, deepcopy

from einops import einsum
import torch.nn as nn

import torchlinops.config as config
from .namedlinop import NamedLinop
from .nameddim import ND, NS, NamedShape, Shape
from .identity import Identity

__all__ = ["Dense"]


class Dense(NamedLinop):
    """
    Example:
    x: [A, Nx, Ny]
    weightshape: [A, T]
    oshape: [T, Nx, Ny]

    shape:
        diag.ioshape = [Nx, Ny]
        dense.ishape = [A]
        dense.oshape = [T]


    # Diagonal/elementwise dimensions
    # Einsum can infer which dims are shared (literally the point of
    # the summation notation)
    x: [C, A, Nx, Ny]
    weight: [C, A, A1]
    oshape: [C, A1, Nx, Ny]
    """

    def __init__(
        self,
        weight: Tensor,
        weightshape: Shape,
        ishape: Shape,
        oshape: Shape,
        broadcast_dims: Optional[list] = None,
    ):
        """
        broadcast_dims : list
            A list of the dimensions of weight that are intended to be broadcasted over the input.
            As such, they are excluded from splitting.
        """
        super().__init__(NS(ishape, oshape))
        self.weight = nn.Parameter(weight, requires_grad=False)
        self._shape.add("weightshape", weightshape)

        broadcast_dims = broadcast_dims if broadcast_dims is not None else []
        self._shape.add("broadcast_dims", broadcast_dims)

    @property
    def weightshape(self):
        return self._shape.lookup("weightshape")

    @property
    def broadcast_dims(self):
        return self._shape.lookup("broadcast_dims")

    @property
    def forward_einstr(self):
        return f"{self.einstr(self.ishape)},{self.einstr(self.weightshape)}->{self.einstr(self.oshape)}"

    @property
    def adj_einstr(self):
        return f"{self.einstr(self.oshape)},{self.einstr(self.weightshape)}->{self.einstr(self.ishape)}"

    @staticmethod
    def einstr(arr):
        """
        tup: Iterable of str-able objects
        """
        return " ".join(str(s) for s in arr)

    def forward(self, x):
        return self.fn(self, x, self.weight)

    @staticmethod
    def fn(linop, x, /, weight):
        return einsum(x, weight, linop.forward_einstr)

    @staticmethod
    def adj_fn(linop, x, /, weight):
        return einsum(x, weight.conj(), linop.adj_einstr)

    @staticmethod
    def normal_fn(linop, x, /, weight):
        return linop.adj_fn(linop.fn(x, weight), weight)

    def adjoint(self):
        adj = copy(self)
        adj.weight = nn.Parameter(
            self.weight.conj(), requires_grad=adj.weight.requires_grad
        )
        adj._shape = adj._shape.H
        adj._update_suffix(adjoint=self._name is not None)
        return adj

    def normal(self, inner=None):
        """
        If no inner, consolidate two Dense's into a single Dense
        ishape: [A B X Y]
        oshape: [C D X Y]
        wshape: [A B C D]

        Needs to become
        ishape: [A B X Y]
        oshape: [A1 B1 X Y]
        wshape: [A B A1 B1]

        New weight is attained as
        einsum(weight.conj(), weight, 'A1 B1 C D, A B C D -> A B A1 B1')

        -----
        ishape: [C A]
        oshape: [C1 A]
        wshape = [C C1]

        Needs to become
        ishape: [C A]
        oshape: [C2 A]
        wshape = [C C2]

        einsum(weight.conj(), weight, 'C1 C2, C C1 -> C C2)


        """
        new_oshape = []
        weight_conj_shape = list(deepcopy(self.weightshape))
        wdiag_shape = []
        wout_shape = []
        win_shape = []
        used_shapes = self.ishape + self.oshape
        shape_updates = {}
        # Make new oshape and weight shape
        # Rules:
        # New weightshape
        #   If dim appears in ishape and weightshape but not oshape -> increment
        #   If dim appears in ishape and weightshape AND oshape -> don't increment
        #   If dim doesn't appear in ishape or weightshape -> don't add it to new weightshape
        # Other rules:
        # new ishape is same as old ishape
        # new oshape is ishape but updated with new dimensions
        for dim in self.ishape:
            if dim in self.weightshape:
                if dim not in self.oshape:
                    win_shape.append(dim)
                    new_dim = dim.next_unused(used_shapes)
                    shape_updates[dim] = new_dim
                    wout_shape.append(new_dim)
                else:
                    wdiag_shape.append(dim)
                    new_dim = dim
                i = weight_conj_shape.index(dim)
                weight_conj_shape[i] = new_dim
            else:
                new_dim = dim
            new_oshape.append(new_dim)

        if config.inner_not_relevant(inner):
            # Consolidate dense and dense adjoint into single dense
            new_weight_shape = wdiag_shape + wout_shape + win_shape
            einstr = shapes2einstr(
                self.weightshape,
                weight_conj_shape,
                new_weight_shape,
            )
            new_weight = einsum(self.weight, self.weight.conj(), einstr)
            normal = type(self)(
                new_weight,
                tuple(new_weight_shape),
                self.ishape,
                new_oshape,
            )
            normal._name = self._name
            normal._update_suffix(normal=self._name is not None)
            normal._shape_updates = shape_updates
            return normal
        _shape_updates = getattr(inner, "_shape_updates", {})
        _shape_updates.update(shape_updates)
        pre = copy(self)
        pre.oshape = inner.ishape
        post = self.adjoint()  # Copy happens inside adjoint
        post.ishape = inner.oshape
        post.oshape = new_oshape
        normal = post @ inner @ pre
        normal._shape_updates = _shape_updates
        return normal

    def split_forward(self, ibatch, obatch):
        weight = self.split_forward_fn(ibatch, obatch, self.weight)
        out = copy(self)
        out.weight = nn.Parameter(weight, requires_grad=self.weight.requires_grad)
        return out

    def split_forward_fn(self, ibatch, obatch, /, weight):
        weightbatch = [slice(None)] * len(self.weightshape)
        for dim, batch in zip(self.ishape, ibatch):
            if dim in self.weightshape and dim not in self.broadcast_dims:
                weightbatch[self.weightshape.index(dim)] = batch
        for dim, batch in zip(self.oshape, obatch):
            if dim in self.weightshape and dim not in self.broadcast_dims:
                weightbatch[self.weightshape.index(dim)] = batch
        return weight[weightbatch]

    def size(self, dim: str):
        return self.size_fn(dim, self.weight)

    def size_fn(self, dim: str, weight):
        if dim in self.broadcast_dims:
            return None
        if dim in self.weightshape:
            return weight.shape[self.weightshape.index(dim)]
        return None


def shapes2einstr(shape1, shape2, oshape):
    """Takes 3 tuples and produces the corresponding einsum string

    Examples
    --------

    """

    def to_str(shape):
        return " ".join(str(s) for s in shape)

    return f"{to_str(shape1)},{to_str(shape2)}->{to_str(oshape)}"



import torch
import torch.nn as nn

from .nameddim import NS, isequal
from .namedlinop import NamedLinop

__all__ = ["Add"]


class Add(NamedLinop):
    def __init__(self, *linops):
        assert all(
            isequal(linop.ishape, linops[0].ishape) for linop in linops
        ), f"Add: All linops must share same ishape. Found {linops}"
        assert all(
            isequal(linop.oshape, linops[0].oshape) for linop in linops
        ), f"Add: All linops must share same oshape. Linops: {linops}"
        # TODO: specialize the ishape and oshape on most specific one
        super().__init__(NS(linops[0].ishape, linops[0].oshape))
        self.linops = nn.ModuleList(linops)

    def forward(self, x):
        return sum(linop(x) for linop in self.linops)

    @staticmethod
    def fn(linop, x: torch.Tensor, /, data_list):
        assert (
            len(linop.linops) == len(data_list)
        ), f"Length {len(data_list)} data_list does not match length {len(linop.linops)} chain linop"
        return sum(linop.fn(x, *data) for linop, data in zip(linop.linops, data_list))

    @staticmethod
    def adj_fn(linop, x: torch.Tensor, /, data_list):
        assert (
            len(linop.linops) == len(data_list)
        ), f"Length {len(data_list)} data_list does not match length {len(linop.linops)} chain adjoint linop"
        return sum(
            linop.adj_fn(x, *data) for linop, data in zip(linop.linops, data_list)
        )

    @staticmethod
    def normal_fn(linop, x: torch.Tensor, /, data_list):
        # Note: Alternatively, make every possible combination of terms? Might be faster in some cases?
        return linop.adj_fn(linop.fn(x, data_list), data_list)

    def split_forward(self, ibatch, obatch):
        """ibatch, obatch specified according to the shape of the
        forward op
        """
        linops = [linop.split_forward(ibatch, obatch) for linop in self.linops]
        return type(self)(*linops)

    def split_forward_fn(self, ibatch, obatch, data_list):
        """Split data into batches
        ibatches, obatches specified according to the shape of the
        forward op
        """
        data = [
            linop.split_forward_fn(ibatch, obatch, data)
            for linop, data in zip(self.linops, data_list)
        ]
        return data

    def adjoint(self):
        return type(self)(*(linop.adjoint() for linop in self.linops))

    def size(self, dim):
        for linop in self.linops:
            out = linop.size(dim)
            if out is not None:
                return out
        return None

    def size_fn(self, dim, data):
        for linop, data in zip(self.linops, data):
            out = linop.size_fn(dim, data)
            if out is not None:
                return out
        return None

    @property
    def dims(self):
        return set().union(*[linop.dims for linop in self.linops])

    @property
    def H(self):
        """Adjoint operator"""
        if self._adj is None:
            linops = list(linop.adjoint() for linop in self.linops)
            _adj = type(self)(*linops)
            self._adj = [_adj]  # Prevent registration as a submodule
        return self._adj[0]

    @property
    def N(self):
        """Normal operator (is this really necessary?)"""
        if self._normal is None:
            linops = list(linop.normal() for linop in self.linops)
            _normal = type(self)(*linops)
            self._normal = [_normal]  # Prevent registration as a submodule
        return self._normal[0]

    def flatten(self):
        return [self]

    def __getitem__(self, idx):
        return self.linops[idx]

    def __len__(self):
        return len(self.linops)

    def __repr__(self):
        linop_chain = " + ".join(repr(linop) for linop in self.linops)
        return linop_chain



import pytest
from itertools import product

from math import prod

import torch

import torchlinops.functional as F
from torchlinops import ArrayToBlocks

from torchlinops.utils import is_adjoint
from torchlinops.tests.test_base import BaseNamedLinopTests


PYTEST_GPU_MARKS = [
    pytest.mark.gpu,
    pytest.mark.skipif(
        not torch.cuda.is_available(), reason="GPU is required but not available"
    ),
]


class TestArrayToBlocks(BaseNamedLinopTests):
    equality_check = "approx"

    # Fixture/test parameterization
    devices = ["cpu"]
    instances = ["small3d"]

    @pytest.fixture(
        scope="class",
        params=product(instances, devices),
    )
    def linop_input_output(self, request):
        spec, dev = request.param
        spec = request.getfixturevalue(spec)
        device = torch.device(dev)
        dtype = torch.complex64

        batch_size = spec["N"]
        im_size = spec["shape"]
        block_size = spec["block_size"]
        stride = spec["stride"]
        mask = spec["mask"]
        nblocks = F.get_nblocks(im_size, block_size, stride)
        ishape = (*batch_size, *im_size)
        if mask is not None:
            oshape = (*batch_size, *nblocks, int(mask.sum()))
        else:
            oshape = (*batch_size, *nblocks, *block_size)

        linop = ArrayToBlocks(im_size, block_size, stride, mask)
        x = torch.randn(ishape, dtype=dtype, device=device)

        y = torch.randn(oshape, dtype=dtype, device=device)

        return linop, x, y

    @pytest.fixture(scope="class")
    def small3d(self, request):
        spec = {
            "N": (1,),
            "shape": (15, 15, 15),
            "block_size": (3, 3, 3),
            "stride": (1, 1, 1),
            "mask": torch.tensor(
                [
                    [
                        [0, 0, 0],
                        [0, 1, 0],
                        [0, 0, 0],
                    ],
                    [
                        [0, 1, 0],
                        [1, 1, 1],
                        [0, 1, 0],
                    ],
                    [
                        [0, 0, 0],
                        [0, 1, 0],
                        [0, 0, 0],
                    ],
                ],
                dtype=bool,
            ),
        }
        return spec



from typing import Optional
from collections.abc import Iterable

__all__ = ["batch_iterator"]


def batch_iterator(total: int, batch_size: Optional[int]) -> Iterable[tuple[int, int]]:
    assert total > 0, f"batch_iterator called with {total} elements"
    if batch_size is None:
        return [(0, total)]
    delim = list(range(0, total, batch_size)) + [total]
    return zip(delim[:-1], delim[1:])



from torch import Tensor
import torch.fft as fft

__all__ = ["cfft", "cifft", "cfft2", "cifft2", "cfftn", "cifftn"]


def cfftn(x, dim=None, norm="ortho"):
    """Matches Sigpy's fft, but in torch
    c = centered
    """
    x = fft.ifftshift(x, dim=dim)
    x = fft.fftn(x, dim=dim, norm=norm)
    x = fft.fftshift(x, dim=dim)
    return x


def cifftn(x, dim=None, norm="ortho"):
    """Matches Sigpy's fft adjoint, but in torch"""
    x = fft.ifftshift(x, dim=dim)
    x = fft.ifftn(x, dim=dim, norm=norm)
    x = fft.fftshift(x, dim=dim)
    return x


# Convenience functions
def cfft(x: Tensor, **kwargs):
    return cfftn(x, dim=(-1,), **kwargs)


def cifft(x: Tensor, **kwargs):
    return cifftn(x, dim=(-1,), **kwargs)


def cfft2(x: Tensor, **kwargs):
    return cfftn(x, dim=(-2, -1), **kwargs)


def cifft2(x: Tensor, **kwargs):
    return cifftn(x, dim=(-2, -1), **kwargs)



from collections import defaultdict
from copy import deepcopy
from typing import Sequence, Any, Tuple
from pprint import pprint
from warnings import warn

from ._nameddim import ELLIPSES, ANY

__all__ = ["partition", "isequal", "iscompatible"]


def partition(seq: Sequence, val: Any) -> Tuple[Sequence, Sequence, Sequence]:
    """Split a sequence on the first occurence of some value

    Examples
    --------
    >>> partition(("A", "B", "C"), "B")
    (('A',), ('B',), ('C',))

    >>> partition(("A", "B", "C"), "C")
    (('A', 'B'), ('C',), ())

    >>> partition(("A", "B", "C"), "D")
    (('A', 'B', 'C'), (), ())


    """
    if val not in seq:
        return seq, type(seq)(), type(seq)()
    n = seq.index(val)
    first = seq[:n]
    middle = type(seq)((val,))
    last = seq[n + 1 :]
    return first, middle, last


def isequal(
    shape1: Sequence,
    shape2: Sequence,
    return_assignments: bool = False,
) -> bool:
    """Test if two sequences with ellipses are compatible

    Implemented with bottom-up DP

    Parameters
    ----------
    shape1, shape2 : Sequence
        The sequences of tokens to compare.
    ELLIPSES : str, default = "..."
        The wildcard that can match any number of tokens.

    Returns
    -------
    bool
        Whether shape1 and shape2 are compatible.

    Examples
    --------

    >>> isequal(("A", "B"), ("A", "B"))
    True
    >>> isequal(("A", "C"), ("A",))
    False
    >>> isequal(("A", "C"), tuple())
    False
    >>> isequal(("A", "C"), ("...",))
    True
    >>> isequal(("A", "C", "..."), ("...",))
    True
    >>> isequal(("A", "B", "C"), ("A", "...", "C"))
    True
    >>> isequal(("...", "A", "C", "..."), ("...",))
    True
    >>> isequal(("...", "A", "C"), ("B", "C"))
    False

    # Wildcards
    >>> isequal(("A", "B"), ("A", "()"))
    True
    >>> isequal(("A",), ("()", "()"))
    False

    # Think about this one...
    >>> isequal(("...", "A", "C", "..."), ("...", "A"))
    True
    """
    ptrs = [[None for _ in range(len(shape2) + 1)] for _ in range(len(shape1) + 1)]
    # Base cases
    ptrs[0][0] = (0, 0)  # True (note that bool(tuple()) == False)
    for i in range(1, len(shape1) + 1):
        ptrs[i][0] = (-1, 0) if shape1[0] == ELLIPSES else None
    for j in range(1, len(shape2) + 1):
        ptrs[0][j] = (0, -1) if shape2[0] == ELLIPSES else None
    for i in range(1, len(shape1) + 1):
        for j in range(1, len(shape2) + 1):
            if ptrs[i - 1][j - 1]:
                if shape1[i - 1] == ELLIPSES or shape2[j - 1] == ELLIPSES:
                    val = (-1, -1)
                elif shape1[i - 1] == shape2[j - 1]:
                    val = (-1, -1)
                elif shape1[i - 1] == ANY or shape2[j - 1] == ANY:
                    val = (-1, -1)
                else:
                    val = None
            elif ptrs[i - 1][j]:
                if shape2[j - 1] == ELLIPSES:
                    val = (-1, 0)
                else:
                    val = None
            elif ptrs[i][j - 1]:
                if shape1[i - 1] == ELLIPSES:
                    val = (0, -1)
                else:
                    val = None
            else:
                val = None
            ptrs[i][j] = val

    if return_assignments:
        if not ptrs[-1][-1]:
            return False, None
        # Traverse in reverse order
        assignments = defaultdict(list)
        row, col = len(shape1), len(shape2)
        while row > 0 and col > 0:
            assignments[row - 1].append(col - 1)
            drow, dcol = ptrs[row][col]
            row = row + drow
            col = col + dcol
        return True, assignments
    return bool(ptrs[-1][-1])


# def get_assignments(ptrs):
#     """Assign indices from first to second sequence."""
#     len_shape1 = len(ptrs[0]) - 1
#     len_shape2 = len(ptrs) - 1
#     assignment = defaultdict(list)
#     row = 1
#     for col in range(1, len_shape1 + 1):
#         while row < len_shape2 + 1 and ptrs[row][col]:
#             assignment[col - 1].append(row - 1)
#             row += 1
#     return assignment


def iscompatible(
    shape1: Sequence,
    shape2: Sequence,
    return_assignments: bool = False,
):
    """Whether the two shapes are length-compatible.

    >>> iscompatible(("A","B"), ("C", "D"))
    True
    >>> iscompatible(("...",), ("A","B"))
    True
    >>> iscompatible(("B","..."), ("A","..."))
    True
    >>> iscompatible(("...",), tuple())
    True

    """
    if isinstance(shape1, str) or isinstance(shape2, str):
        warn(
            f"Strings detected in iscompatible call - are you sure this is what you want? {shape1}, {shape2}"
        )

    shape1 = [ANY if s != ELLIPSES else ELLIPSES for s in deepcopy(shape1)]
    shape2 = [ANY if s != ELLIPSES else ELLIPSES for s in deepcopy(shape2)]
    return isequal(shape1, shape2, return_assignments=return_assignments)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import pytest

import torch
from torchlinops import Dense, split_linop, create_batched_linop, BatchSpec


def test_split_linop():
    ishape = ("B", "N")
    oshape = ("B", "M")
    B = 10
    M, N = (3, 7)
    weight = torch.randn(B, M, N)
    weightshape = ("B", "M", "N")
    device = "cpu"
    A = Dense(weight, weightshape, ishape, oshape)
    linops, in_slc, out_slc = split_linop(A, {"N": 2, "M": 1})

    # tile indices
    n, m = 1, 2
    # Input
    x_n = torch.randn(B, 2)
    y_m = linops[n, m](x_n)
    # True operator
    A_mn = Dense(weight[:, m : m + 1, 2 * n : 2 * (n + 1)], weightshape, ishape, oshape)
    y_m_ref = A_mn(x_n)
    assert torch.allclose(y_m, y_m_ref)


def test_create_batched_linop():
    ishape = ("B", "N")
    oshape = ("B", "M")
    B = 10
    M, N = (3, 7)
    weight = torch.randn(B, M, N)
    weightshape = ("B", "M", "N")
    device = "cpu"
    A = Dense(weight, weightshape, ishape, oshape)

    Abatch = create_batched_linop(A, BatchSpec(dict(N=2, M=1)))
    x = torch.randn(B, N)
    assert Abatch(x).allclose(A(x), rtol=1e-3)


@pytest.mark.gpu
@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="GPU is required but not available"
)
def test_create_batched_linop_multi_device():
    ishape = ("B", "N")
    oshape = ("B", "M")
    B = 10
    M, N = (3, 7)
    weight = torch.randn(B, M, N)
    weightshape = ("B", "M", "N")
    device = "cpu"
    A = Dense(weight, weightshape, ishape, oshape)

    Abatch = create_batched_linop(
        A,
        [
            BatchSpec(
                dict(N=2),
                device_matrix=[torch.device("cpu"), torch.device("cuda:0")],
                base_device=torch.device("cpu"),
            ),
            BatchSpec(dict(M=1)),
        ],
    )
    x = torch.randn(B, N)
    assert Abatch(x).allclose(A(x), rtol=1e-3)



__all__ = ["default_to"]


def default_to(*vals, typecast: bool = False):
    """Get the first non-None value, right to left order.

    Most "default" value goes first.
    """
    if len(vals) == 0:
        return None
    typecls = type(vals[0])
    if len(vals) == 1:
        return vals[0]
    for val in reversed(vals):
        if val is not None:
            if typecast:
                return typecls(val)
            return val



import pytest
from itertools import product

from math import prod

import torch

from torchlinops import Interpolate
from torchlinops.functional._interp.tests._valid_pts import get_valid_locs

from torchlinops.utils import is_adjoint
from torchlinops.tests.test_base import BaseNamedLinopTests


PYTEST_GPU_MARKS = [
    pytest.mark.gpu,
    pytest.mark.skipif(
        not torch.cuda.is_available(), reason="GPU is required but not available"
    ),
]


class TestInterp(BaseNamedLinopTests):
    equality_check = "approx"

    # Fixture/test parameterization
    devices = ["cpu"]
    instances = ["small3d"]

    @pytest.fixture(
        scope="class",
        params=product(instances, devices),
    )
    def linop_input_output(self, request):
        spec, dev = request.param
        spec = request.getfixturevalue(spec)
        device = torch.device(dev)
        dtype = torch.complex64

        width = spec["width"]
        grid_size = spec["grid_size"]
        locs_batch_size = spec["locs_batch_size"]
        ndim = len(grid_size)
        npts = prod(locs_batch_size)
        batch_size = spec["N"]
        ishape = (*batch_size, *grid_size)
        oshape = (*batch_size, *locs_batch_size)
        locs = get_valid_locs(locs_batch_size, grid_size, ndim, width, device)

        linop = Interpolate(locs, grid_size, width=width, kernel="kaiser_bessel")
        x = torch.randn(ishape, dtype=dtype, device=device)
        y = torch.randn(oshape, dtype=dtype, device=device)

        return linop, x, y

    @pytest.fixture(scope="class")
    def small3d(self, request):
        N = (1, 1)
        grid_size = (10, 8, 9)
        locs_batch_size = (3, 5)
        width = 4.0

        spec = {
            "N": N,
            "grid_size": grid_size,
            "locs_batch_size": locs_batch_size,
            "width": width,
        }
        return spec


def test_interp_slc():
    locs = torch.rand(8, 6, 5, 3)
    grid_size = (5, 5, 5)
    locs_batch_shape = ("A", "B", "C")
    linop = Interpolate(locs, grid_size, locs_batch_shape=locs_batch_shape)
    tile = {"B": slice(0, 4), "C": slice(2, 3)}
    obatch = [tile.get(dim, slice(None)) for dim in locs_batch_shape]
    linop_split = linop.split(linop, tile)
    assert linop_split.locs.shape == (8, 4, 1, 3)
    assert (linop_split.locs == linop.locs[obatch]).all()



import torch

from torchlinops.utils.benchmark import benchmark_and_summarize
import torchlinops.functional as F


def main():
    device = torch.device("cuda:0")
    B = 15
    Nx, Ny, Nz = (220, 220, 220)
    grid_size = (Nx, Ny, Nz)
    # ishape = (B, *grid_size)
    oshape = (48, 500, 1600)

    def gen_data():
        vals = torch.randn(*oshape, device=device)
        idx = [torch.randint(0, N, oshape, device=device) for N in grid_size]
        return vals, idx

    # Define random test functions
    def index_index_put():
        vals, idx = gen_data()
        # idx = torch.stack(idx, dim=-1)
        # return multi_grid(vals, idx, final_size=grid_size)
        return F.index_adjoint(vals, idx, grid_size=grid_size)

    def index_interp0():
        vals, idx = gen_data()
        locs = torch.stack(idx, dim=-1).float()
        return F.grid(vals, locs, grid_size=grid_size, width=1.0, kernel="spline")

    def index_multi_grid():
        vals, idx = gen_data()
        idx = torch.stack(idx, dim=-1)
        return multi_grid(vals, idx, final_size=grid_size)

    benchmark_and_summarize(gen_data, name="gen_data", num_iters=100)
    benchmark_and_summarize(
        index_interp0, name="grid with kernel=spline and width=0", num_iters=100
    )
    benchmark_and_summarize(index_index_put, name="index_put_", num_iters=100)
    benchmark_and_summarize(index_multi_grid, name="multi_grid", num_iters=100)

    # Test for correctness
    breakpoint()
    vals, idx = gen_data()
    y1 = F.index_adjoint(vals, idx, grid_size=grid_size)
    locs = torch.stack(idx, dim=-1).float()
    y2 = F.grid(vals, locs, grid_size=grid_size, width=1.0, kernel="spline")
    idx = torch.stack(idx, dim=-1)
    y3 = multi_grid(vals, idx, final_size=grid_size)

    assert y1.allclose(y2)
    assert y2.allclose(y3)


def multi_grid(
    x: torch.Tensor, idx: torch.Tensor, final_size: tuple, raveled: bool = False
):
    """Grid values in x to im_size with indices given in idx
    x: [N... I...]
    idx: [I... ndims] or [I...] if raveled=True
    raveled: Whether the idx still needs to be raveled or not

    Returns:
    Tensor with shape [N... final_size]

    Notes:
    Adjoint of multi_index
    Might need nonnegative indices
    """
    if not raveled:
        assert (
            len(final_size) == idx.shape[-1]
        ), f"final_size should be of dimension {idx.shape[-1]}"
        idx = ravel(idx, final_size, dim=-1)
    ndims = len(idx.shape)
    assert (
        x.shape[-ndims:] == idx.shape
    ), f"x and idx should correspond in last {ndims} dimensions"
    x_flat = torch.flatten(x, start_dim=-ndims, end_dim=-1)  # [N... (I...)]
    idx_flat = torch.flatten(idx)

    batch_dims = x_flat.shape[:-1]
    y = torch.zeros(
        (*batch_dims, *final_size), dtype=x_flat.dtype, device=x_flat.device
    )
    y = y.reshape((*batch_dims, -1))
    y = y.index_add_(-1, idx_flat, x_flat)
    y = y.reshape(*batch_dims, *final_size)
    return y


def ravel(x: torch.Tensor, shape: tuple, dim: int):
    """
    x: torch.LongTensor, arbitrary shape,
    shape: Shape of the array that x indexes into
    dim: dimension of x that is the "indexing" dimension

    Returns:
    torch.LongTensor of same shape as x but with indexing dimension removed
    """
    out = 0
    shape_shifted = tuple(shape[1:]) + (1,)
    for s, s_next, i in zip(shape, shape_shifted, range(x.shape[dim])):
        out += torch.select(x, dim, i) % s
        out *= s_next
    return out


if __name__ == "__main__":
    main()



from math import ceil
import torch


def get_valid_locs(
    locs_batch_size,
    grid_size,
    ndim,
    width,
    device,
    centered: bool = False,
    valid: bool = True,
):
    """Avoid circular padding weirdness by sampling valid locations only
    Allow non-valid locations by setting valid=False
    """
    out = []
    for d in range(ndim):
        if valid:
            lower = ceil(width / 2)
            upper = grid_size[d] - 1 - lower
        else:
            lower = 0
            upper = grid_size[d] - 1
        locs = torch.rand(*locs_batch_size, device=device)
        locs = locs * (upper - lower) + lower
        out.append(locs)
    out = torch.stack(out, dim=-1).contiguous()
    if centered:
        sz = torch.tensor(grid_size)
        out -= sz / 2
    return out



from typing import Optional
from jaxtyping import Shaped, Bool
from torch import Tensor

from itertools import product

import torch

try:
    import triton
    import triton.language as tl
    from .casting import scalar_cast as cast

    TRITON_ENABLED = True
except ImportError:
    from torchlinops.utils import fake_triton as triton, fake_tl as tl

    TRITON_ENABLED = False


from .nblocks import get_nblocks

__all__ = ["fold"]

# Maximum block sizes in each direction
# Determined by CUDA
MAX_1D_BLOCK_SIZE = 2**20
MAX_2D_BLOCK_SIZE = 2**10
MAX_3D_BLOCK_SIZE = 2**6


def fold(
    x,
    im_size: tuple,
    block_size: tuple,
    stride: tuple,
    mask: Optional[Bool[Tensor, "..."]] = None,
    output: Optional[Tensor] = None,
) -> Tensor:
    """Accumulate an array of blocks into a full array

    Parameters
    ----------
    x : Tensor
        Shape [B..., blocks, block_size]

    Returns
    -------
    Tensor: Shape [B..., *im_size]
        If mask is not None, block_size will be an int equal to the number of True elements in the mask
        Otherwise it will be the full block shape.
    """
    x_flat, shapes, is_complex = _prep_fold(x, im_size, block_size, stride, mask)

    if is_complex:
        x_flat = torch.view_as_real(x_flat)
        x_flat = torch.flatten(x_flat, -2, -1)  # Flatten real/imag into last dim
    y_flat = _fold(x_flat, output=output, **shapes)
    y = y_flat.reshape(*shapes["batch_shape"], *shapes["im_size"])
    if is_complex:
        y = y.reshape(*y.shape[:-1], y.shape[-1] // 2, 2)
        y = torch.view_as_complex(y)
    return y


def _fold(
    x: Shaped[Tensor, "B ..."],
    block_size: tuple[int, ...],
    stride: tuple[int, ...],
    ndim: int,
    im_size: tuple[int, ...],
    nblocks: tuple[int, ...],
    nbatch: int,
    output: Optional[Tensor] = None,
    **kwargs,
):
    """Implementation of fold"""
    # Check dtype if output buffer is provided
    if output is not None:
        if not output.dtype == x.dtype:
            raise ValueError(
                f"Output and input dtypes must match but got output {output.dtype} != input {x.dtype}"
            )

    if x.shape[-2 * ndim :] != (*nblocks, *block_size):
        raise RuntimeError(
            f"Fold expected input with full size {(*nblocks, *block_size)} but got {x.shape}"
        )
    if x.is_cuda and ndim in FOLD.keys():
        x = x.contiguous()  # Ensure contiguity
        with torch.cuda.device(x.device):
            if output is None:
                # Allocate output
                y = torch.zeros(
                    nbatch,
                    *im_size,
                    device=x.device,
                    dtype=x.dtype,
                )
            else:
                # Use existing buffer
                y = output.reshape(nbatch, *im_size).zero_()
            grid = _get_grid(ndim, nbatch, im_size)

            FOLD[ndim][grid](
                x,
                y,
                nbatch,
                *nblocks,
                *block_size,
                *im_size,
                *stride,
            )
    else:
        y = _fold_torch(
            x, block_size, stride, ndim, im_size, nblocks, nbatch, out=output
        )
    return y


def _get_grid(ndim: int, nbatch, im_size):
    if ndim == 1:
        grid = lambda meta: (  # noqa: E731
            nbatch * triton.cdiv(im_size[0], meta["X_BLOCK_SIZE"]),
        )
    elif ndim == 2:
        grid = lambda meta: (  # noqa: E731
            nbatch * triton.cdiv(im_size[0], meta["X_BLOCK_SIZE"]),
            triton.cdiv(im_size[1], meta["Y_BLOCK_SIZE"]),
        )
    elif ndim == 3:
        grid = lambda meta: (  # noqa: E731
            nbatch * triton.cdiv(im_size[0], meta["X_BLOCK_SIZE"]),
            triton.cdiv(im_size[1], meta["Y_BLOCK_SIZE"]),
            triton.cdiv(im_size[2], meta["Z_BLOCK_SIZE"]),
        )
    else:
        raise ValueError(f"Invalid ndim = {ndim}")
    return grid


@triton.heuristics(
    values={
        "X_BLOCK_SIZE": lambda args: min(
            MAX_1D_BLOCK_SIZE,
            triton.next_power_of_2(args["x_stride"]),
        ),
    },
)
@triton.jit
def _fold1d(
    in_ptr,
    out_ptr,
    # Number of batches
    nbatch: int,
    # Number of blocks
    x_nblocks: int,
    # Size of each block
    x_block_dim: int,
    # Size of the input data
    x_size: int,
    # Stride of the blocks
    x_stride: int,
    # Size of the triton block (power of 2)
    X_BLOCK_SIZE: tl.constexpr,
):
    dtype = in_ptr.type.element_ty
    pid_0 = tl.program_id(0)
    x_blocks_per_batch = cdiv(x_size, X_BLOCK_SIZE)

    # Batch index, Block index
    N, Ix = pid_0 // x_blocks_per_batch, pid_0 % x_blocks_per_batch

    # Convert types
    x_nblocks = cast(x_nblocks, tl.int64)
    x_block_dim = cast(x_block_dim, tl.int64)
    x_size = cast(x_size, tl.int64)
    x_stride = cast(x_stride, tl.int32)

    nblocks = x_nblocks
    block_dim = x_block_dim
    size = x_size

    in_offset = N * nblocks * block_dim

    # Find overlapping blocks with range
    x_lower = Ix * X_BLOCK_SIZE
    x_upper = x_lower + X_BLOCK_SIZE
    Bx_lower = cdiv(x_lower - x_block_dim + 1, x_stride)
    Bx_upper = cdiv(x_upper, x_stride)  # non-inclusive

    # Initialize output
    output = tl.zeros((1, X_BLOCK_SIZE), dtype)
    x_range = tl.arange(0, X_BLOCK_SIZE) + x_lower
    x_mask = x_range < x_size
    out_offset = N * size
    out_range = x_range
    out_mask = x_mask

    out_range = out_range[None]
    out_mask = out_mask[None]

    for Bx in range(Bx_lower, Bx_upper):
        if Bx >= 0 and Bx < x_nblocks:
            x_Lpad = Bx * x_stride - x_lower
            x_in_range = tl.arange(0, X_BLOCK_SIZE)
            x_in_mask = ((x_in_range - x_Lpad) >= 0) & (
                (x_in_range - x_Lpad) < x_block_dim
            )
            # Load block
            block_offset = Bx * block_dim
            block_offset = block_offset - x_Lpad
            in_range = x_in_range
            in_mask = x_in_mask
            blk = tl.load(in_ptr + in_offset + block_offset + in_range, in_mask)
            output += blk
    tl.store(out_ptr + out_offset + out_range, output, out_mask)


@triton.heuristics(
    values={
        "X_BLOCK_SIZE": lambda args: min(
            MAX_2D_BLOCK_SIZE,
            triton.next_power_of_2(args["x_stride"]),
        ),
        "Y_BLOCK_SIZE": lambda args: min(
            MAX_2D_BLOCK_SIZE,
            triton.next_power_of_2(args["y_stride"]),
        ),
    },
)
@triton.jit
def _fold2d(
    in_ptr,
    out_ptr,
    # Number of batches
    nbatch: int,
    # Number of blocks
    x_nblocks: int,
    y_nblocks: int,
    # Size of each block
    x_block_dim: int,
    y_block_dim: int,
    # Size of the input data
    x_size: int,
    y_size: int,
    # Stride of the blocks
    x_stride: int,
    y_stride: int,
    # Size of the triton block (power of 2)
    X_BLOCK_SIZE: tl.constexpr,
    Y_BLOCK_SIZE: tl.constexpr,
):
    dtype = in_ptr.type.element_ty
    pid_0 = tl.program_id(0)
    pid_1 = tl.program_id(1)
    # x_blocks_per_batch = tl.ceil(x_size / X_BLOCK_SIZE)
    x_blocks_per_batch = cdiv(x_size, X_BLOCK_SIZE)
    y_blocks_per_batch = cdiv(y_size, Y_BLOCK_SIZE)

    # Batch index, Block index
    N, Ix = pid_0 // x_blocks_per_batch, pid_0 % x_blocks_per_batch
    Iy = pid_1 % y_blocks_per_batch

    # Convert types
    x_nblocks = cast(x_nblocks, tl.int64)
    y_nblocks = cast(y_nblocks, tl.int64)
    x_block_dim = cast(x_block_dim, tl.int64)
    y_block_dim = cast(y_block_dim, tl.int64)
    x_size = cast(x_size, tl.int64)
    y_size = cast(y_size, tl.int64)
    x_stride = cast(x_stride, tl.int32)
    y_stride = cast(y_stride, tl.int32)

    nblocks = x_nblocks * y_nblocks
    block_dim = x_block_dim * y_block_dim
    size = x_size * y_size
    in_offset = N * nblocks * block_dim

    # Find overlapping blocks with range
    x_lower = Ix * X_BLOCK_SIZE
    x_upper = x_lower + X_BLOCK_SIZE
    Bx_lower = cdiv(x_lower - x_block_dim + 1, x_stride)
    Bx_upper = cdiv(x_upper, x_stride)  # non-inclusive
    y_lower = Iy * Y_BLOCK_SIZE
    y_upper = y_lower + Y_BLOCK_SIZE
    By_lower = cdiv(y_lower - y_block_dim + 1, y_stride)
    By_upper = cdiv(y_upper, y_stride)  # non-inclusive

    # Initialize output
    output = tl.zeros((1, X_BLOCK_SIZE, Y_BLOCK_SIZE), dtype)
    x_range = tl.arange(0, X_BLOCK_SIZE) + x_lower
    x_mask = x_range < x_size
    y_range = tl.arange(0, Y_BLOCK_SIZE) + y_lower
    y_mask = y_range < y_size

    out_offset = N * size
    out_range = x_range[:, None] * y_size + y_range[None, :]
    out_mask = x_mask[:, None] & y_mask[None, :]

    out_range = out_range[None]
    out_mask = out_mask[None]

    for Bx in range(Bx_lower, Bx_upper):
        if Bx >= 0 and Bx < x_nblocks:
            x_Lpad = Bx * x_stride - x_lower
            x_in_range = tl.arange(0, X_BLOCK_SIZE)
            x_in_mask = ((x_in_range - x_Lpad) >= 0) & (
                (x_in_range - x_Lpad) < x_block_dim
            )
            for By in range(By_lower, By_upper):
                if By >= 0 and By < y_nblocks:
                    y_Lpad = By * y_stride - y_lower
                    y_in_range = tl.arange(0, Y_BLOCK_SIZE)
                    y_in_mask = ((y_in_range - y_Lpad) >= 0) & (
                        (y_in_range - y_Lpad) < y_block_dim
                    )
                    # Load block
                    block_offset = Bx * y_nblocks * block_dim + By * block_dim
                    block_offset = block_offset - (x_Lpad * y_block_dim + y_Lpad)
                    in_range = x_in_range[:, None] * y_block_dim + y_in_range[None, :]
                    in_mask = x_in_mask[:, None] & y_in_mask[None, :]
                    blk = tl.load(in_ptr + in_offset + block_offset + in_range, in_mask)
                    output += blk
    tl.store(out_ptr + out_offset + out_range, output, out_mask)


@triton.heuristics(
    values={
        "X_BLOCK_SIZE": lambda args: min(
            MAX_3D_BLOCK_SIZE,
            triton.next_power_of_2(args["x_stride"]),
        ),
        "Y_BLOCK_SIZE": lambda args: min(
            MAX_3D_BLOCK_SIZE,
            triton.next_power_of_2(args["y_stride"]),
        ),
        "Z_BLOCK_SIZE": lambda args: min(
            MAX_3D_BLOCK_SIZE,
            triton.next_power_of_2(args["z_stride"]),
        ),
    },
)
@triton.jit
def _fold3d(
    in_ptr,
    out_ptr,
    # Number of batches
    nbatch: int,
    # Number of blocks
    x_nblocks: int,
    y_nblocks: int,
    z_nblocks: int,
    # Size of each block
    x_block_dim: int,
    y_block_dim: int,
    z_block_dim: int,
    # Size of the input data
    x_size: int,
    y_size: int,
    z_size: int,
    # Stride of the blocks
    x_stride: int,
    y_stride: int,
    z_stride: int,
    # Size of the triton block (power of 2)
    X_BLOCK_SIZE: tl.constexpr,
    Y_BLOCK_SIZE: tl.constexpr,
    Z_BLOCK_SIZE: tl.constexpr,
):
    dtype = in_ptr.type.element_ty
    pid_0 = tl.program_id(0)
    pid_1 = tl.program_id(1)
    pid_2 = tl.program_id(2)
    x_blocks_per_batch = cdiv(x_size, X_BLOCK_SIZE)
    y_blocks_per_batch = cdiv(y_size, Y_BLOCK_SIZE)
    z_blocks_per_batch = cdiv(z_size, Z_BLOCK_SIZE)

    # Batch index, Block index
    N, Ix = pid_0 // x_blocks_per_batch, pid_0 % x_blocks_per_batch
    Iy = pid_1 % y_blocks_per_batch
    Iz = pid_2 % z_blocks_per_batch

    # Convert types
    x_nblocks = cast(x_nblocks, tl.int64)
    y_nblocks = cast(y_nblocks, tl.int64)
    z_nblocks = cast(z_nblocks, tl.int64)
    x_block_dim = cast(x_block_dim, tl.int64)
    y_block_dim = cast(y_block_dim, tl.int64)
    z_block_dim = cast(z_block_dim, tl.int64)
    x_size = cast(x_size, tl.int64)
    y_size = cast(y_size, tl.int64)
    z_size = cast(z_size, tl.int64)
    x_stride = cast(x_stride, tl.int32)
    y_stride = cast(y_stride, tl.int32)
    z_stride = cast(z_stride, tl.int32)

    nblocks = x_nblocks * y_nblocks * z_nblocks
    block_dim = x_block_dim * y_block_dim * z_block_dim
    size = x_size * y_size * z_size

    in_offset = N * nblocks * block_dim

    # Find overlapping blocks with range
    x_lower = Ix * X_BLOCK_SIZE
    x_upper = x_lower + X_BLOCK_SIZE
    Bx_lower = cdiv(x_lower - x_block_dim + 1, x_stride)
    Bx_upper = cdiv(x_upper, x_stride)  # non-inclusive
    y_lower = Iy * Y_BLOCK_SIZE
    y_upper = y_lower + Y_BLOCK_SIZE
    By_lower = cdiv(y_lower - y_block_dim + 1, y_stride)
    By_upper = cdiv(y_upper, y_stride)  # non-inclusive
    z_lower = Iz * Z_BLOCK_SIZE
    z_upper = z_lower + Z_BLOCK_SIZE
    Bz_lower = cdiv(z_lower - z_block_dim + 1, z_stride)
    Bz_upper = cdiv(z_upper, z_stride)  # non-inclusive

    # Initialize output
    output = tl.zeros((1, X_BLOCK_SIZE, Y_BLOCK_SIZE, Z_BLOCK_SIZE), dtype)
    x_range = tl.arange(0, X_BLOCK_SIZE) + x_lower
    x_mask = x_range < x_size
    y_range = tl.arange(0, Y_BLOCK_SIZE) + y_lower
    y_mask = y_range < y_size
    z_range = tl.arange(0, Z_BLOCK_SIZE) + z_lower
    z_mask = z_range < z_size

    out_offset = N * size
    out_range = (
        x_range[:, None, None] * y_size + y_range[None, :, None]
    ) * z_size + z_range[None, None, :]
    out_mask = x_mask[:, None, None] & (y_mask[None, :, None] & z_mask[None, None, :])

    out_range = out_range[None]
    out_mask = out_mask[None]

    for Bx in range(Bx_lower, Bx_upper):
        if Bx >= 0 and Bx < x_nblocks:
            x_Lpad = Bx * x_stride - x_lower
            x_in_range = tl.arange(0, X_BLOCK_SIZE)
            x_in_mask = ((x_in_range - x_Lpad) >= 0) & (
                (x_in_range - x_Lpad) < x_block_dim
            )
            for By in range(By_lower, By_upper):
                if By >= 0 and By < y_nblocks:
                    y_Lpad = By * y_stride - y_lower
                    y_in_range = tl.arange(0, Y_BLOCK_SIZE)
                    y_in_mask = ((y_in_range - y_Lpad) >= 0) & (
                        (y_in_range - y_Lpad) < y_block_dim
                    )
                    for Bz in range(Bz_lower, Bz_upper):
                        if Bz >= 0 and Bz < z_nblocks:
                            z_Lpad = Bz * z_stride - z_lower
                            z_in_range = tl.arange(0, Z_BLOCK_SIZE)
                            z_in_mask = ((z_in_range - z_Lpad) >= 0) & (
                                (z_in_range - z_Lpad) < z_block_dim
                            )
                            # Load block
                            block_offset = (
                                (Bx * y_nblocks + By) * z_nblocks + Bz
                            ) * block_dim
                            block_offset = block_offset - (
                                (x_Lpad * y_block_dim + y_Lpad) * z_block_dim + z_Lpad
                            )
                            in_range = (
                                x_in_range[:, None, None] * y_block_dim
                                + y_in_range[None, :, None]
                            ) * z_block_dim + z_in_range[None, None, :]
                            in_mask = x_in_mask[:, None, None] & (
                                y_in_mask[None, :, None] & z_in_mask[None, None, :]
                            )
                            blk = tl.load(
                                in_ptr + in_offset + block_offset + in_range, in_mask
                            )
                            output += blk
    tl.store(out_ptr + out_offset + out_range, output, out_mask)


@triton.jit
def cdiv(a, b):
    return tl.cast(tl.ceil(a / b), tl.int32)


if TRITON_ENABLED:
    FOLD = {1: _fold1d, 2: _fold2d, 3: _fold3d}
else:
    FOLD = {}


def _fold_torch(
    x: Shaped[Tensor, "B ..."],
    block_size: tuple[int, ...],
    stride: tuple[int, ...],
    ndim: int,
    im_size: tuple[int, ...],
    nblocks: tuple[int, ...],
    nbatch: int,
    out: Optional[Tensor] = None,
) -> Shaped[Tensor, "B I ..."]:
    """Fallback option

    Note: Compile takes forever
    """
    if out is None:
        out = torch.zeros((nbatch, *im_size), device=x.device, dtype=x.dtype)
    else:
        out = out.reshape(nbatch, *im_size).zero_()
    # Python implementation
    for batch in range(nbatch):
        for blk in product(*(range(nblk) for nblk in nblocks)):
            blk_slc = tuple(
                slice(iblk * st, iblk * st + blk_sz)
                for iblk, st, blk_sz in zip(blk, stride, block_size)
            )
            in_idx = (batch, *blk)
            out_idx = (batch, *blk_slc)
            out[out_idx] += x[in_idx]
    return out


def _prep_fold(x, im_size, block_size, stride, mask):
    # Handle mask
    if mask is not None:
        nblocks = get_nblocks(im_size, block_size, stride)
        if len(x.shape) > len(nblocks) + 1:
            x_batch_shape = x.shape[: -(len(nblocks) + 1)]
        else:
            x_batch_shape = []
        tmp = torch.zeros(
            *x_batch_shape,
            *nblocks,
            *block_size,
            dtype=x.dtype,
            device=x.device,
        )
        tmp[..., mask] = x
        x = tmp
        mask = mask.to(x.device)
    is_complex = torch.is_complex(x)
    ndim = len(block_size)
    stride = stride if stride is not None else (1,) * ndim
    nblocks = get_nblocks(im_size, block_size, stride)
    if any(b == 0 for b in nblocks):
        raise ValueError(
            f"Found 0 in nblocks: {nblocks} with im_size {im_size}, block_size {block_size} and stride {stride} - make sure there is at least one block in each direction."
        )
    if is_complex:
        im_size = list(im_size)
        im_size[-1] *= 2
        block_size = list(block_size)
        block_size[-1] *= 2
        stride = list(stride)
        stride[-1] *= 2

    # Add or infer batch dim
    batch_shape = x.shape[: (-2 * ndim)]
    if 2 * ndim < len(x.shape):
        x_flat = x.flatten(0, len(x.shape) - (2 * ndim) - 1)
    else:
        x_flat = x[None]
    nbatch = x_flat.shape[0]

    return (
        x_flat,
        {
            "ndim": ndim,
            "im_size": im_size,
            "stride": stride,
            "nblocks": nblocks,
            "nbatch": nbatch,
            "batch_shape": batch_shape,
            "mask": mask,
            "block_size": block_size,
        },
        is_complex,
    )



import pytest

from copy import copy

import torch

from torchlinops import Diagonal

from torchlinops.utils import inner, is_adjoint


@pytest.fixture
def A():
    M = 10
    weight = torch.randn(M, 1, 1, dtype=torch.complex64)
    # weightshape = ("M",)
    ioshape = ("M", "N", "P")
    A = Diagonal(weight, ioshape)
    return A


def test_diagonal(A):
    M = A.weight.shape[0]
    N, P = 5, 7
    x = torch.randn(M, N, P, dtype=torch.complex64)
    y = torch.randn(M, N, P, dtype=torch.complex64)
    assert is_adjoint(A, x, y)


# Specific tests
@pytest.mark.xfail  # Deprecated behavior: changing a non-() dim to ()
def test_diagonal_shape_renaming(A):
    B = copy(A)

    # Change to () first
    new_ioshape = ("()", "N1", "P1")
    B.oshape = new_ioshape
    assert B.ishape == new_ioshape

    # Change to something else after
    new_ioshape = ("M1", "N1", "P1")
    B.oshape = new_ioshape
    assert B.ishape == new_ioshape



from dataclasses import dataclass
from functools import partial
from math import ceil
from typing import Literal, Optional
from warnings import warn

import numpy as np
import torch
from torchlinops.utils import NDList, batch_iterator, dict_product

from .add import Add
from .concat import Concat
from .device import ToDevice
from .nameddim import ND
from .namedlinop import NamedLinop

__all__ = ["split_linop", "create_batched_linop", "BatchSpec"]

Batch = tuple[int, slice]
# Represents a single batch at index 0 over the full extent
# Could convert to a full class
DEFAULT_BATCH = (0, slice(None))
Tile = dict[ND | str, Batch]


@dataclass
class BatchSpec:
    batch_sizes: dict[ND | str, int]
    device_matrix: Optional[np.ndarray | list] = None
    base_device: Optional[torch.device] = "cpu"

    def __post_init__(self):
        if not isinstance(self.batch_sizes, dict):
            warn(
                f"Got {self.batch_sizes} of type {type(self.batch_sizes).__name__} for batch_sizes instead of dict."
            )

    def broadcast_device_matrix(self, linop):
        if self.device_matrix is not None:
            batch_dims = list(self.batch_sizes.keys())
            sizes = {dim: linop.size(dim) for dim in linop.dims}
            # Create and broadcast device_matrix over requested split
            tiled_shape = tuple(
                ceil(sizes[dim] / self.batch_sizes[dim]) for dim in batch_dims
            )
            if not isinstance(self.device_matrix, np.ndarray):
                device_matrix = np.array(self.device_matrix, dtype=object)
            device_matrix = fuzzy_broadcast_to(device_matrix, tiled_shape)
            return device_matrix
        return None

    @staticmethod
    def linop_to_device(linop, device, base_device, memory_aware: bool = True):
        linop = (
            ToDevice(device, base_device, linop.oshape)
            @ linop.to(device, memory_aware=memory_aware)
            @ ToDevice(base_device, device, linop.ishape)
        )
        return linop


def create_batched_linop(linop, batch_specs: BatchSpec | list[BatchSpec]):
    """
    Examples
    --------
    >>> from torchlinops import Dense
    >>> non_gpu_batchspec = BatchSpec({"B": 2})
    >>> gpu_batchspec = BatchSpec({"C": 1}, device_matrix=[torch.device("cuda:0"), "cpu"])

    """
    if isinstance(batch_specs, BatchSpec):
        # Ensure list
        batch_specs = [batch_specs]

    if len(batch_specs) == 0:
        return linop
    batch_spec = batch_specs[0]
    linops, _, _ = split_linop(linop, batch_spec.batch_sizes)
    device_matrix = batch_spec.broadcast_device_matrix(linop)
    if device_matrix is not None:
        if device_matrix.shape != linops.shape:
            raise ValueError(
                f"Broadcasted device matrix with shape {device_matrix.shape} can't be broadcasted to exact shape of linop tiles with shape {linops.shape}"
            )
        device_matrix = device_matrix.reshape(-1)  # Flatten

    # Work with flattened linop
    linops_shape = linops.shape
    linops = linops.reshape(-1)  # Flatten
    for i, linop in enumerate(linops):
        tiled_linop = create_batched_linop(linop, batch_specs[1:])
        if device_matrix is not None:
            device = device_matrix[i]
            tiled_linop = batch_spec.linop_to_device(
                tiled_linop,
                device,
                batch_spec.base_device,
                memory_aware=True,
            )
        linops[i] = tiled_linop
    linops = linops.reshape(linops_shape)

    for dim in reversed(batch_spec.batch_sizes):
        # Manual axis reduction because I made Concat and Add too nice
        flat_linops = linops.reshape(-1, linops.shape[-1])
        new_linops = np.empty(flat_linops.shape[0], dtype=object)
        for i, linop_arr in enumerate(flat_linops):
            linop = linop_arr[0]
            if dim in linop.ishape and dim in linop.oshape:
                new_linop = Concat(*linop_arr, idim=dim, odim=dim)
            elif dim not in linop.ishape and dim in linop.oshape:
                new_linop = Concat(*linop_arr, odim=dim)
            elif dim in linop.ishape and dim not in linop.oshape:
                new_linop = Concat(*linop_arr, idim=dim)
            else:
                new_linop = Add(*linop_arr)
            new_linops[i] = new_linop
        linops = new_linops.reshape(linops.shape[:-1])
    return linops.item()


def split_linop(linop: NamedLinop, batch_sizes: dict[ND | str, int]):
    """Split a linop into smaller linops according to some batch sizes

    Parameters
    ----------
    linop : NamedLinop
        The NamedLinop to be split.
    batch_sizes : dict[ND | str, int]
        Dictionary mapping dims to batch sizes for those dims.
    device_matrix : list | np.ndarray, optional
        Optional list of devices to broadcast the linop over. See notes for device
        broadcasting rules.

    Returns
    -------

    """
    # Precompute sizes and shapes
    batch_sizes = {ND.infer(k): v for k, v in batch_sizes.items()}
    sizes = {dim: linop.size(dim) for dim in linop.dims}

    # Make tiles. Each tile is a dictionary mapping a dimension to an integer
    # index of the tile and a slice over that dimension.
    batch_iterators = make_batch_iterators(sizes, batch_sizes)
    tiles: list[dict[ND, Batch]] = list(dict_product(batch_iterators))

    # Allocate outputs
    batch_dims = list(batch_sizes.keys())
    tiled_shape = tuple(ceil(sizes[dim] / batch_sizes[dim]) for dim in batch_dims)
    linops = np.ndarray(tiled_shape, dtype=object)
    input_batches = np.ndarray(tiled_shape, dtype=object)
    output_batches = np.ndarray(tiled_shape, dtype=object)

    for tile in tiles:
        idx = _tile_get_idx(tile, batch_dims)
        linop_tile = split_linop_with_tile(linop, tile)
        linop_flat = linop_tile.flatten()
        first_linop, last_linop = linop_flat[0], linop_flat[-1]
        linops[idx] = linop_tile
        input_batches[idx] = [
            tile.get(dim, DEFAULT_BATCH)[1] for dim in first_linop.ishape
        ]
        output_batches[idx] = [
            tile.get(dim, DEFAULT_BATCH)[1] for dim in last_linop.oshape
        ]
    return linops, input_batches, output_batches


def fuzzy_broadcast_to(arr: np.ndarray, target_shape):
    """Broadcast an array to a target shape, truncating or repeating if necessary.

    Examples
    --------
    >>> import numpy as np
    >>> arr = np.array([1, 2, 3])
    >>> fuzzy_broadcast_to(arr, (4,))
    array([1, 2, 3, 1])
    >>> fuzzy_broadcast_to(arr, (3, 1))
    array([[1, 2, 3],
           [1, 2, 3],
           [1, 2, 3]])
    >>> arr2 = np.array([1, 2])
    >>> fuzzy_broadcast_to(arr2, (3, 5))
    array([[1, 2, 1, 2, 1],
           [1, 2, 1, 2, 1],
           [1, 2, 1, 2, 1]])
    """
    # Ensure target shape and arr.shape have same length
    if len(target_shape) < arr.ndim:
        target_shape = (1,) * (arr.ndim - len(target_shape)) + tuple(target_shape)
    while len(target_shape) > arr.ndim:
        arr = np.expand_dims(arr, 0)

    if not is_broadcastable(arr.shape, target_shape):
        # Identify offending dimensions and repeat
        repeats = []
        for source_dim, target_dim in zip(arr.shape, target_shape):
            if source_dim == target_dim or source_dim == 1:
                repeats.append(1)
            elif source_dim < target_dim:
                repeats.append(int(ceil(target_dim / source_dim)))
            else:
                repeats.append(1)
        arr = tile_along_axes(arr, repeats)
        # Trim the excess
        slices = tuple(slice(0, dim) for dim in target_shape)
        arr = arr[slices]
    target_shape = np.broadcast_shapes(target_shape, arr.shape)
    return np.broadcast_to(arr, target_shape)


def split_linop_with_tile(linop: NamedLinop, tile: Tile):
    """Split a linop according to batch specified in tile"""
    slice_map = {key: value[1] for key, value in tile.items()}
    linop_tile = linop.split(linop, slice_map)
    return linop_tile


def _tile_get_idx(tile: Tile, batch_dims) -> tuple[int]:
    """Get all indices from the tile"""
    return tuple(tile.get(dim, DEFAULT_BATCH)[0] for dim in batch_dims)


def make_batch_iterators(
    total_sizes: dict[str, int],
    batch_sizes: dict[str, int],
) -> Tile:
    """Construct dictionaries mapping batchable dims to lists of slices
    corresponding to the actual batches

    Also includes an int index at dim 0

    Explanation
    -----------
    If we have batch size 3 for dim D (i.e. batch_sizes = {"D": 3})
    and the total size for dim D is 7, then

    batch_iterators["D"] = [(0, slice(0, 3)), (1, slice(3, 6)), (2, slice(6, 7))]

    If "E" is some other dimension not batched, then

    batch_iterators["E"] = [(0, slice(None))]



    """
    batch_iterators = {}
    for dim, total in total_sizes.items():
        batch_iterators[dim] = (
            [
                (i, slice(a, b))
                for i, (a, b) in enumerate(batch_iterator(total, batch_sizes[dim]))
            ]
            if dim in batch_sizes
            else [(0, slice(None))]
        )
    return batch_iterators


def flatten_recursive(nested_list, max_depth: Optional[int] = None):
    """Flatten a nested list, optionally to a maximum depth

    Examples
    -------
    >>> flatten_recursive([[1, 2], [3, 4], [[5, 6]]])
    [1, 2, 3, 4, 5, 6]

    # Setting max_depth = 1 will avoid flattening the last list completely
    >>> flatten_recursive([[1, 2], [3, 4], [[5, 6]]], max_depth=1)
    [1, 2, 3, 4, [5, 6]]

    """
    flat_list = []
    for item in nested_list:
        if isinstance(item, list):
            if max_depth is None:
                flat_list.extend(flatten_recursive(item))
            elif max_depth > 0:
                flat_list.extend(flatten_recursive(item, max_depth - 1))
            else:
                flat_list.append(item)
        else:
            flat_list.append(item)
    return flat_list


def is_broadcastable(a, b):
    try:
        np.broadcast_shapes(a, b)
        return True
    except ValueError:
        return False


def repeat_along_axes(arr, repeats):
    """
    Repeat a numpy array along its axes.

    Parameters:
    - arr: np.ndarray
    - repeats: list or tuple of ints, with one repeat count per axis.
               If len(repeats) < arr.ndim, remaining axes are not repeated.
               If len(repeats) > arr.ndim, array is reshaped to add new axes.

    Returns:
    - Repeated array.
    """
    arr = np.asarray(arr)
    ndim = arr.ndim
    if len(repeats) > ndim:
        # Add singleton dimensions to match the number of repeats
        arr = arr.reshape(arr.shape + (1,) * (len(repeats) - ndim))
    elif len(repeats) < ndim:
        repeats = list(repeats) + [1] * (ndim - len(repeats))

    for axis, repeat in enumerate(repeats):
        if repeat > 1:
            arr = np.repeat(arr, repeat, axis=axis)
    return arr


def tile_along_axes(arr, tile_counts):
    """
    Tile a numpy array along its axes.

    Parameters:
    - arr: np.ndarray
    - tile_counts: list or tuple of ints, with one tile count per axis.
                   If len(tile_counts) < arr.ndim, missing axes default to 1.
                   If len(tile_counts) > arr.ndim, singleton dimensions are added.

    Returns:
    - Tiled array.
    """
    arr = np.asarray(arr)
    ndim = arr.ndim
    if len(tile_counts) < ndim:
        tile_counts = list(tile_counts) + [1] * (ndim - len(tile_counts))
    elif len(tile_counts) > ndim:
        arr = arr.reshape(arr.shape + (1,) * (len(tile_counts) - ndim))

    return np.tile(arr, tile_counts)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import pytest

import torch

from torchlinops.functional._interp._circ_pad import circular_pad, circular_pad_adjoint
from torchlinops.utils import inner


def test_circular_pad_2d():
    ishape = (15, 20)
    x = torch.randn(*ishape)
    padding = [4, 4, 2, 2]

    y = circular_pad(x, padding)
    assert y.shape == (19, 28)


def test_circular_pad_nd():
    ishape = (5, 15, 20, 19, 18)
    x = torch.randn(*ishape)
    padding = [1, 1, 2, 2, 3, 3, 5, 5]
    y = circular_pad(x, padding)
    assert y.shape == (5, 25, 26, 23, 20)


def test_circular_pad_adjoint_1d():
    ishape = (3,)
    padding = [1, 1]
    oshape = list(ishape)
    for i, (padleft, padright) in enumerate(zip(padding[::2], padding[1::2])):
        oshape[-(i + 1)] += padleft + padright
    x = torch.randn(*ishape, dtype=torch.complex64)
    y = torch.randn(*oshape, dtype=torch.complex64)

    Ax = circular_pad(x, padding)
    AHy = circular_pad_adjoint(y, padding)

    assert torch.allclose(inner(x, AHy), inner(y, Ax).conj())


def test_circular_pad_adjoint_2d():
    ishape = (3, 3)
    padding = [2, 2, 1, 1]
    oshape = list(ishape)
    for i, (padleft, padright) in enumerate(zip(padding[::2], padding[1::2])):
        oshape[-(i + 1)] += padleft + padright
    x = torch.randn(*ishape, dtype=torch.complex64)
    y = torch.randn(*oshape, dtype=torch.complex64)

    Ax = circular_pad(x, padding)
    AHy = circular_pad_adjoint(y, padding)

    assert torch.allclose(inner(x, AHy), inner(y, Ax).conj())



from collections import defaultdict
from copy import deepcopy

import torch
import torch.nn as nn
from torchlinops import BatchSpec, Dense, Diagonal, Dim, create_batched_linop
from torchlinops.utils import MemReporter


def main():
    """Requires CUDA"""
    device = torch.device("cuda:0")
    B, C, X, Y = 6, 5, 64, 64
    input_ = torch.randn(X, Y)
    S = Dense(torch.randn(C, X, Y), Dim("CXY"), ishape=Dim("XY"), oshape=Dim("CXY"))
    F = Dense(torch.randn(B, 1, X, Y), Dim("B"), ishape=Dim("CXY"), oshape=Dim("BCXY"))

    # Make linop
    A = F @ S
    A.to(device)
    print(A)
    print("Not batched (GPU)")
    MemReporter().report(A)
    A = create_batched_linop(A, [BatchSpec({"C": 1}), BatchSpec({"B": 2})])
    print(A)

    # Print memory usage
    print("Batched (GPU)")
    MemReporter().report(A)

    # Serialize
    torch.save(A, "A.pt")
    A2 = torch.load("A.pt", weights_only=False)

    # Print memory usage
    print("Deserialized (GPU -> GPU)")
    MemReporter().report(A2)

    # Deepcopy
    # Memory expands?
    A_copy = deepcopy(A)
    print("A_copy")
    MemReporter().report(A_copy)
    breakpoint()
    print("A_copy is different: ")
    breakpoint()

    # CPU
    # Memory drastically expands again...
    A_copy = deepcopy(A)
    MemReporter().report(A_copy)
    A3 = A_copy.to("cpu")
    print("CPU")
    MemReporter().report(A3)

    # Preserve references
    A4 = A.to("cpu")
    print("CPU Attempt #2")
    MemReporter().report(A4)


def test_deepcopy():
    """Requires CUDA"""
    device = torch.device("cuda:0")
    B, C, X, Y = 6, 5, 64, 64
    input_ = torch.randn(X, Y)
    S = Dense(torch.randn(C, X, Y), Dim("CXY"), ishape=Dim("XY"), oshape=Dim("CXY"))
    F = Dense(torch.randn(B, 1, X, Y), Dim("B"), ishape=Dim("CXY"), oshape=Dim("BCXY"))

    # Make linop
    A = F @ S
    A.to(device)
    A = create_batched_linop(A, [BatchSpec({"C": 1}), BatchSpec({"B": 2})])

    # deepcopy
    # Memory drastically expands again...
    A_copy = deepcopy(A)
    MemReporter().report(A_copy)

    assert id(A[0][0][0].weight) != id(A_copy[0][0][0].weight)


if __name__ == "__main__":
    main()
    test_deepcopy()



from typing import Callable

import torch

__all__ = ["inner", "is_adjoint"]


def inner(x, y):
    """Complex inner product"""
    return torch.sum(x.conj() * y)


def is_adjoint(
    A: Callable,
    x: torch.Tensor,
    y: torch.Tensor,
    atol: float = 1e-5,
    rtol: float = 1e-8,
):
    """
    The adjoint test states that if A and AH are adjoints, then
    inner(y, Ax) = inner(AHy, x)
    """
    yAx = inner(y, A(x))
    xAHy = inner(A.H(y), x)
    return torch.isclose(yAx, xAHy, atol=atol, rtol=rtol).all()



import logging
from pathlib import Path
import shutil
import sys
from typing import Optional, Literal
import yaml

logger = logging.getLogger(__name__)

__all__ = [
    "mkdir",
    "save_yaml_options",
]


def mkdir(root_path, action: Optional[Literal["o", "l", "a"]] = None):
    """
    root_path should be a pathlib.Path
    """
    if root_path.exists():
        # Overwrite?
        if action is not None:
            opt = action
        else:
            opt = input(
                f"{root_path} exists. [O]verwrite/[L]eave/[A]bort? [o/l/a] "
            ).lower()
        if opt == "o":
            logger.warning(f"Overwriting {root_path}.")
            shutil.rmtree(root_path)
        elif opt == "l":
            logger.debug(f"Saving to {root_path}")
        else:
            logger.debug("Aborting.")
            sys.exit()
    Path(root_path).mkdir(parents=True, exist_ok=True)


def save_yaml_options(config_file: Path, opt):
    def noop(self, *args, **kwargs):
        pass

    yaml.emitter.Emitter.process_tag = noop
    with open(config_file, "w") as f:
        yaml.dump(opt, f)



from typing import Optional
from jaxtyping import Integer
from torch import Tensor

from copy import copy

import torch
import torch.nn as nn

from .namedlinop import NamedLinop
from .add import Add
from .identity import Zero
from .nameddim import NS, isequal, ELLIPSES, NDorStr, ND

from torchlinops.utils import default_to, INDENT

__all__ = ["Concat"]


class Concat(NamedLinop):
    """Concatenate some linops along an existing dimension

    Linops need not output tensors of the same size, but they should
    output tensors of the same number of dimensions

    Stacking type depends on dimensions provided

    Horizontal stacking
    stacking along an input dimension:

    A B C

    Vertical stacking
    stacking along an output dimension:

    A
    B
    C

    Diagonal stacking:
    stacking along a separate input and output dimensions

    A . .
    . B .
    . . C


    """

    def __init__(
        self,
        *linops,
        idim: Optional[NDorStr] = None,
        odim: Optional[NDorStr] = None,
    ):
        self._check_linop_compatibility(linops)
        super().__init__(NS(linops[0].ishape, linops[0].oshape))
        self.linops = nn.ModuleList(list(linops))

        # Handle ishape
        ishape = linops[0].ishape
        if idim is not None:
            self.idim = ND.infer(idim)
            if any(linop.size(self.idim) is None for linop in linops):
                raise ValueError(
                    f"Found linop with undefined size for dim {self.idim} when attempting concat."
                )
            self.isizes = [linop.size(self.idim) for linop in linops]
            self.islices = torch.tensor(self.isizes).cumsum(0)  # Keep on CPU
            # self.islices = nn.Parameter(islices, requires_grad=False)
        else:
            self.idim = None
            self.isizes = None
            self.islices = None

        # Handle oshape
        oshape = linops[0].oshape
        if odim is not None:
            self.odim = ND.infer(odim)
            if any(linop.size(self.odim) is None for linop in linops):
                raise ValueError(
                    f"Found linop with undefined size for dim {self.odim} when attempting concat."
                )
            self.osizes = [linop.size(self.odim) for linop in linops]
            self.oslices = torch.tensor(self.osizes).cumsum(0)  # Keep on CPU
            # self.oslices = nn.Parameter(oslices, requires_grad=False)
        else:
            self.odim = None
            self.osizes = None
            self.oslices = None

        self.idim_idx = self._infer_dim_idx(self.idim, ishape)
        self.odim_idx = self._infer_dim_idx(self.odim, oshape)

    def forward(self, x):
        return self.fn(self, x)

    @staticmethod
    def fn(concat, x):
        return concat._fn(
            x,
            concat.linops,
            concat.idim_idx,
            concat.odim_idx,
            concat.islices,
            concat.oslices,
        )

    @staticmethod
    def adj_fn(concat, x):
        adj_linops = [linop.H for linop in concat.linops]
        return concat._fn(
            x,
            adj_linops,
            concat.odim_idx,
            concat.idim_idx,
            concat.oslices,
            concat.islices,
        )

    @staticmethod
    def _fn(x: Tensor, linops, idim_idx, odim_idx, islices, oslices):
        """Unifies forward and adjoint functionality for stacked linops"""
        # Split inputs
        if idim_idx is not None:  # Diagonal, Horizontal
            # if sum(sizes) != x.shape[idim_idx]:
            if islices[-1] != x.shape[idim_idx]:
                raise ValueError(
                    f"Concat Linop expecting input of size {islices[-1]} got input of size {x.shape} with non-matching concat size {x.shape[idim_idx]}"
                )
            xs = x.tensor_split(islices, dim=idim_idx)[:-1]  # Omit final slice
        else:  # Vertical
            xs = [x] * len(oslices)

        # Compute linop(x) for all xs
        if odim_idx is not None:  # Diagonal, Vertical
            ys = []
            for xi, linop in zip(xs, linops):
                ys.append(linop(xi))
            return torch.concatenate(ys, dim=odim_idx)

        # Memory savings by accumulating
        # Horizontal
        y = 0.0
        # Combine outputs
        for xi, linop in zip(xs, linops):
            y += linop(xi)
        return y

    @staticmethod
    def normal_fn(concat, x):
        return concat.adj_fn(concat, concat.fn(concat, x))

    def size(self, dim):
        return self.size_fn(dim)

    def size_fn(self, dim, /):
        if dim == self.idim:
            return sum(self.isizes)
        elif dim == self.odim:
            return sum(self.osizes)
        else:
            return self.linops[0].size(dim)

    def split_forward(self, ibatch, obatch):
        """Split concat linop, making a new concat linop if necessary"""
        ibatches = self.subslice(ibatch, self.idim_idx, self.islices, len(self.linops))
        obatches = self.subslice(obatch, self.odim_idx, self.oslices, len(self.linops))

        output_linop_idxs = ibatches.keys() & obatches.keys()
        output_linop_idxs = sorted(list(output_linop_idxs))
        if len(output_linop_idxs) == 0:
            # No linops satisfy this slice (diagonal stacking)
            return Zero(self.ishape, self.oshape)
        elif len(output_linop_idxs) == 1:
            # Singleton linop
            linop_idx = output_linop_idxs.pop()
            linop = self.linops[linop_idx]
            ibatch, obatch = ibatches[linop_idx], obatches[linop_idx]
            return linop.split_forward(ibatch, obatch)
        else:
            output_linop_idxs = sorted(list(output_linop_idxs))
            output_linops = []
            for i in output_linop_idxs:
                linop = self.linops[i]
                ibatch, obatch = ibatches[i], obatches[i]
                output_linops.append(linop.split_forward(ibatch, obatch))
            return type(self)(*output_linops, idim=self.idim, odim=self.odim)

    def split_forward_fn(self, ibatch, obatch, data_list):
        """Split concat linop, making a new concat linop if necessary

        Parameters
        ----------
        data_list : list, same length as linops
            List of data for each linop in this concat linop

        """
        ibatches = self.subslice(ibatch, self.idim_idx, self.islices, len(self.linops))
        obatches = self.subslice(obatch, self.odim_idx, self.oslices, len(self.linops))

        output_linop_idxs = ibatches.keys() & obatches.keys()
        if len(output_linop_idxs) == 0:
            # No linops satisfy this slice (diagonal stacking)
            return 0.0  # TODO is this ok
        elif len(output_linop_idxs) == 1:
            # Singleton linop
            linop_idx = output_linop_idxs.pop()
            linop = self.linops[linop_idx]
            data = data_list[linop_idx]
            ibatch, obatch = ibatches[linop_idx], obatches[linop_idx]
            return linop.split_forward_fn(ibatch, obatch, data)
        else:
            output_linop_idxs = sorted(list(output_linop_idxs))
            output_linop_data = []
            for i in output_linop_idxs:
                linop = self.linops[i]
                data = data_list[i]
                ibatch, obatch = ibatches[i], obatches[i]
                output_linop_data.append(linop.split_forward_fn(ibatch, obatch, data))
            return output_linop_data

    @staticmethod
    def subslice(batch: list[slice], dim_idx: Optional[int], slices, num_linops):
        """Given a slice over some dims of a concat linop,
        return a mapping from the linop index to the relevant sub-slice for that linop.
        """
        linops_batch = {}
        if dim_idx is not None:
            slc = batch[dim_idx]
            slice_partition = slices.detach().cpu().numpy().tolist()
            slice_partition.insert(0, 0)
            sub_linop_slices = partition_slices(slice_partition, slc)
            for i, slc in sub_linop_slices:
                sub_linop_batch = copy(batch)
                sub_linop_batch[dim_idx] = slc
                linops_batch[i] = sub_linop_batch
        else:
            for i in range(num_linops):
                linops_batch[i] = batch
        return linops_batch

    def adjoint(self):
        adj_linops = [linop.H for linop in self.linops]
        return type(self)(*adj_linops, idim=self.odim, odim=self.idim)

    def normal(self, inner=None):
        if inner is None:
            if self.idim is None:  # Vertical (inner product)
                return Add(*(linop.N for linop in self.linops))
            elif self.odim is None:  # Horizontal (outer product)
                new_idim, new_odim = self._get_new_normal_io_dims(
                    self.linops[0].shape, self.idim
                )
                rows = []
                new_shape = self.linops[0].shape.N
                for linop_left in self.linops:
                    row = []
                    for linop_right in self.linops:
                        if linop_left == linop_right:
                            new_linop = linop_right.N
                        else:
                            new_linop = linop_left.H @ linop_right
                            new_linop.shape = new_shape
                        row.append(new_linop)
                    row = type(self)(*row, idim=new_idim, odim=None)
                    rows.append(row)
                return type(self)(*rows, idim=None, odim=new_odim)
            else:  # Diagonal
                diag = []
                new_idim, new_odim = self._get_new_normal_io_dims(
                    self.linops[0].shape, self.idim
                )
                for linop in self.linops:
                    diag.append(linop.N)
                return type(self)(*diag, idim=new_idim, odim=new_odim)
        return super().normal(inner)

    @staticmethod
    def _get_new_normal_io_dims(shape, dim) -> tuple:
        new_shape = shape.N
        i = new_shape.ishape.index(dim)
        new_idim = new_shape.ishape[i]
        new_odim = new_shape.oshape[i]
        return new_idim, new_odim

    @staticmethod
    def _check_linop_compatibility(linops: list[NamedLinop]):
        """Ensure linops can actually be concatenated along the requested dimension"""
        target_shape = linops[0].shape
        for linop in linops:
            if not (
                isequal(target_shape.ishape, linop.ishape)
                and isequal(target_shape.oshape, linop.oshape)
            ):
                raise ValueError(
                    f"Incompatible linops being stacked. Target shape: {target_shape} but got linop shape: {linop.shape}"
                )

    @staticmethod
    def _infer_dim_idx(dim: ND, shape: tuple[ND, ...]) -> int:
        """Get index of dim within requested shape tuple

        Tries to infer index in the presence of ellipses "..." shapes
        Returns positive int if possible
        Otherwise, tries to return negative int
        Fails if neither is possible.

        """
        if dim is None:
            return None
        if dim not in shape:
            raise ValueError(
                f"Provided concat dimension {dim} not found in shape {shape}"
            )
        shape_list = [str(s) for s in shape]
        dim_idx = shape_list.index(str(dim))
        pre, post = shape_list[:dim_idx], shape_list[dim_idx + 1 :]
        if ELLIPSES in pre:
            if ELLIPSES in post:
                raise ValueError(
                    f"Cannot infer concat dimension for dim {dim} from shape {shape}"
                )
            else:
                return -(len(post) + 1)
        return len(pre)

    def __getitem__(self, idx):
        linops = self.linops[idx]
        if isinstance(linops, NamedLinop):
            return linops
        return type(self)(*linops, idim=self.idim, odim=self.odim)

    def __len__(self):
        return len(self.linops)

    def __repr__(self):
        output = ""
        output += INDENT.indent(self.repr_name + f"({self._shape}\n")
        with INDENT:
            for linop in self.linops:
                output += repr(linop) + "\n"
            output += INDENT.indent(f"idim = {self.idim}, odim = {self.odim}\n")
        output += INDENT.indent(")")
        return output


def get_slice_start_stop(slice_obj, array_length) -> tuple[int, int]:
    """Calculate the start and stop indices of a slice object for an array of a given length.

    Parameters
    ----------
    slice_obj : slice
        The slice object to interpret.
    array_length : int
        The length of the array being indexed.

    Returns
    -------
    tuple of int
        A tuple containing the start and stop indices as integers.

    Examples
    --------
    >>> s = slice(-3, 7)
    >>> length = 10
    >>> get_slice_start_stop(s, length)
    (7, 7)

    >>> s = slice(None, -2)
    >>> length = 5
    >>> get_slice_start_stop(s, length)
    (0, 3)
    """
    # Handle default values for start and stop
    start = slice_obj.start if slice_obj.start is not None else 0
    stop = slice_obj.stop if slice_obj.stop is not None else array_length

    # Adjust negative indices
    if start < 0:
        start += array_length
    if stop < 0:
        stop += array_length

    # Clamp values to be within array bounds
    start = max(0, min(start, array_length))
    stop = max(0, min(stop, array_length))

    return start, stop


def partition_slices(partition, slc):
    """Generate sub-slices for a given slice and partition.

    Parameters
    ----------
    partition : list
        A list of integers representing the partition boundaries [0, a1, a2, ..., ak, N].
        Includes 0 and ending
    slc : slice
        A slice object representing the range [start, stop).

    Returns
    -------
    list
        A list of tuples
            First entry is the index of the interval
            Second entry is sub-slice object corresponding to that interval.

    Examples
    --------
    >>> partition_slices([0, 5, 10, 15, 20], slice(3, 6))
    [(0, slice(3, 5, None)), (1, slice(0, 1, None))]

    >>> partition_slices([0, 5, 10, 15, 20], slice(3, 11))
    [(0, slice(3, 5, None)), (1, slice(0, 5, None)), (2, slice(0, 1, None))]

    """
    # Validate the partition
    if partition[0] != 0:
        raise ValueError("The first boundary in the partition must be 0.")
    if not all(partition[i] <= partition[i + 1] for i in range(len(partition) - 1)):
        raise ValueError("Partition boundaries must be non-decreasing.")
    if slc.stop is not None and slc.stop > partition[-1]:
        raise ValueError(
            f"Slice {slc} is out of range for array of length {partition[-1]}"
        )
    if slc.step is not None and slc.step != 1:
        raise NotImplementedError(
            f"Partition slicing with step != 1 is not currently supported."
        )

    start, stop = get_slice_start_stop(slc, partition[-1])
    result = []

    for i in range(len(partition) - 1):
        # Interval boundaries
        interval_start = partition[i]
        interval_end = partition[i + 1]

        # Check overlap of slice with the current interval
        if interval_end <= start:  # Slice starts after the interval
            continue
        if interval_start >= stop:  # Slice ends before the interval
            break

        # Calculate sub-slice within the current interval
        sub_start = max(start, interval_start)
        sub_stop = min(stop, interval_end)
        result.append((i, slice(sub_start - interval_start, sub_stop - interval_start)))

    return result


def dynamic_slice(t: Tensor, dim: int, slc: slice | int):
    """Access a tensor at dimension dim with slice slc

    Examples
    --------
    >>> dynamic_slice(torch.randn(2, 3, 4), 2, slc=slice(0, 3)).shape
    torch.Size([2, 3, 3])

    >>> dynamic_slice(torch.randn(2, 3, 4), 2, slc=1).shape
    torch.Size([2, 3])

    """
    if isinstance(slc, int):
        slc = torch.tensor(slc)
    indices = [slice(None)] * t.ndim
    indices[dim] = slc
    return t[indices]
